"""
This module contains the definitions of the class codes, using a custom class.
"""


from collections import OrderedDict


def _is_digit(value):
    if not ((value is None) or (isinstance(value, (float, int)) and 0 <= value <= 100)):
        raise ValueError("Invalid numeric value: {}, type: {}".format(value, type(value)))
    return True


def _is_boolean(value):
    if value not in (True, False, None):
        raise ValueError("Invalid boolean value: {}, type: {}".format(value, type(value)))
    return True


class _ClassCode:

    """Container for the class codes ."""

    def __init__(self, code):

        self.__code = code
        self.__definition = None
        self.__ref_multi = None
        self.__pred_multi = None
        self.__nucl_f1, self.__nucl_prec, self.__nucl_rec = None, None, None
        self.__junc_f1, self.__junc_prec, self.__junc_rec = None, None, None
        self.__reverse = None
        self.__category = None

    def __eq__(self, other):
        if hasattr(other, "code") and self.code == other.code:
            return True
        else:
            return False

    def __hash__(self):
        return hash(self.code)

    def __str__(self):
        lines = list()
        lines.append("- Code: {}".format(self.code))
        lines.append("- Definition: {}".format(self.definition))
        lines.append("- Reference multiexonic: {}".format(self.ref_multi))
        lines.append("- Prediction multiexonic: {}".format(self.pred_multi))
        lines.append("- Nucleotide recall, precision, F1: {}".format(self.nucl))
        lines.append("- Junction recall, precision, F1: {}".format(self.junc))
        lines.append("- Reverse class code: {}".format(self.reverse))
        lines.append("- Category: {}".format(self.category))
        return "\n".join(lines)

    @property
    def code(self):
        return self.__code

    @property
    def _nucl_prec(self):
        return self.__nucl_prec

    @_nucl_prec.setter
    def _nucl_prec(self, value):
        self.__nucl_prec = value

    @property
    def _nucl_rec(self):
        return self.__nucl_rec

    @_nucl_rec.setter
    def _nucl_rec(self, value):
        self.__nucl_rec = value

    @property
    def _nucl_f1(self):
        return self.__nucl_f1

    @_nucl_f1.setter
    def _nucl_f1(self, value):
        self.__nucl_f1 = value

    @property
    def nucl(self):
        if self._nucl_f1 is None and self._nucl_rec is None and self._nucl_prec is None:
            return "NA"
        else:
            line = []
            for val in (self._nucl_rec, self._nucl_prec, self._nucl_f1):
                if val is not None:
                    line.append("{}%".format(val))
                else:
                    line.append("NA")
            return ", ".join(line)

    @property
    def junc(self):
        if self._junc_f1 is None and self._junc_rec is None and self._junc_prec is None:
            return "NA"
        else:
            line = []
            for val in (self._junc_rec, self._junc_prec, self._junc_f1):
                if val is not None:
                    line.append("{}%".format(val))
                else:
                    line.append("NA")
            return ", ".join(line)

    @property
    def _junc_prec(self):
        return self.__junc_prec

    @_junc_prec.setter
    def _junc_prec(self, value):
        self.__junc_prec = value

    @property
    def _junc_rec(self):
        return self.__junc_rec

    @_junc_rec.setter
    def _junc_rec(self, value):
        self.__junc_rec = value

    @property
    def _junc_f1(self):
        return self.__junc_f1

    @_junc_f1.setter
    def _junc_f1(self, value):
        self.__junc_f1 = value

    @property
    def pred_multi(self):
        if self.__pred_multi is None:
            return "NA"
        else:
            return self.__pred_multi

    @pred_multi.setter
    def pred_multi(self, value):
        if _is_boolean(value):
            self.__pred_multi = value

    @property
    def ref_multi(self):
        if self.__ref_multi is None:
            return "NA"
        else:
            return self.__ref_multi

    @ref_multi.setter
    def ref_multi(self, value):
        if _is_boolean(value):
            self.__ref_multi = value

    @property
    def definition(self):
        if self.__definition is None:
            return "NA"
        else:
            return self.__definition

    @definition.setter
    def definition(self, value):
        if isinstance(value, bytes):
            value = value.decode()
        elif not (isinstance(value, str) or value is None):
            raise ValueError("Invalid value for definition: {}, type {}".format(value, type(value)))
        self.__definition = value

    @property
    def category(self):
        if self.__category is None:
            return "NA"
        else:
            return self.__category

    @category.setter
    def category(self, value):
        if isinstance(value, bytes):
            value = value.decode()
        elif not (isinstance(value, str) or value is None):
            raise ValueError("Invalid value for category: {}, type {}".format(value, type(value)))
        self.__category = value

    @property
    def reverse(self):
        if self.__reverse is None:
            return "NA"
        else:
            return self.__reverse

    @reverse.setter
    def reverse(self, value):
        if isinstance(value, bytes):
            value = value.decode()
        elif not (isinstance(value, str) or value is None):
            raise ValueError("Invalid value for reverse: {}, type {}".format(value, type(value)))

        self.__reverse = value


class Equal:

    _code = _ClassCode("=")
    _code.definition = "Complete intron chain match."
    _code.pred_multi, _code.ref_multi = True, True
    _code._junc_f1, _code._junc_prec, _code._junc_rec = [100] * 3
    _code.reverse = "="
    _code.category = "Match"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class UnderScore:
    
    _code = _ClassCode("_")
    _code.definition = "Complete match between two monoexonic transcripts."
    _code.ref_multi, _code.pred_multi = False, False
    _code._nucl_f1 = ">=80"
    _code.reverse = "_"
    _code.category = "Match"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)
    
    
class CodeN:

    _code = _ClassCode("n")
    _code.definition = """Intron chain extension, ie. both transcripts are multiexonic and
    the prediction has novel splice sites outside of the reference transcript boundaries."""
    _code.ref_multi, _code.pred_multi = True, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = (100, "< 100", "<100")
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = (100, "< 100", "<100")
    _code.reverse = "c"
    _code.category = "Extension"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeCapitalJ:
    
    _code = _ClassCode("J")
    _code.definition = """Intron chain extension, ie. both transcripts are multiexonic and
    the prediction has novel splice sites inside of the reference transcript boundaries."""
    _code.ref_multi, _code.pred_multi = True, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = (100, "<= 100", "<100")
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = (100, "< 100", "<100")
    _code.reverse = "C"
    _code.category = "Extension"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeC:
    _code = _ClassCode("c")
    _code.definition = """The prediction is either multiexonic and with its intron chain completely contained
    within that of the reference, or monoexonic and contained within one of the reference exons."""
    _code.pred_multi, _code.ref_multi = None, None
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = "< 100", "100", None
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = "< 100", "100", None
    _code.reverse = "n"
    _code.category = "Extension"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeCapitalC:

    _code = _ClassCode("C")
    _code.definition = """The prediction intron chain is completely contained within that of the reference
    transcript, but it partially debords either into its introns or outside of the reference boundaries."""
    _code.pred_multi, _code.ref_multi = True, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = "<= 100", "< 100", "< 100"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = "< 100", "100", "< 100"
    _code.reverse = "J or j"
    _code.category = "Extension"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeJ:
    _code = _ClassCode("j")
    _code.definition = """Alternative splicing event."""
    _code.ref_multi, _code.pred_multi = True, True
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = "<= 100", "100", "< 100"
    _code.reverse = "j or C"
    _code.category = "Alternative splicing"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeH:
    _code = _ClassCode("h")
    _code.definition = """Structural match between two models where where no splice site is conserved but at least
    one intron of the reference and one intron of the prediction partially overlap."""
    _code.ref_multi, _code.pred_multi = True, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = "> 0", "> 0", "> 0"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "h"
    _code.category = "Alternative splicing"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeG:

    _code = _ClassCode("g")
    _code.definition = """The monoexonic prediction overlaps one or more exons of the reference
     transcript; the borders of the prediction cannot fall inside the introns of the reference.
     The prediction transcript can bridge multiple exons of the reference model."""
    _code.ref_multi, _code.pred_multi = True, False
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = "> 0", "> 0", "0% < F1 < 100"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "G"
    _code.category = "Alternative splicing"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeCapitalG:
    _code = _ClassCode("G")
    _code.definition = """Generic match of a multiexonic prediction transcript versus a monoexonic reference."""
    _code.ref_multi, _code.pred_multi = False, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = "> 0", "> 0", "0% < F1 < 100"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "g"
    _code.category = "Alternative splicing"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeO:
    _code = _ClassCode("o")
    _code.definition = """Generic overlap between two multiexonic transcripts,
    which do not share any overlap among their introns."""
    _code.ref_multi, _code.pred_multi = True, True
    _code.ref_multi, _code.pred_multi = True, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = "> 0", "> 0", "0% < F1 < 100"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "o"
    _code.category = "Overlap"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeE:
    _code = _ClassCode("e")
    _code.definition = """Single exon transcript overlapping one reference exon and at least 10 bps of a
    reference intron, indicating a possible pre-mRNA fragment."""
    _code.ref_multi, _code.pred_multi = True, False
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = "> 0", "> 0", "0% < F1 < 100"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "G"
    _code.category = "Overlap"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeM:
    _code = _ClassCode("m")
    _code.definition = """Generic match between two monoexonic transcripts."""
    _code.ref_multi, _code.pred_multi = False, False
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = None, None, "< 80"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = None, None, None
    _code.reverse = "m"
    _code.category = "Overlap"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeI:
    _code = _ClassCode("i")
    _code.definition = "Monoexonic prediction completely contained within one intron of the reference transcript."
    _code.ref_multi, _code.pred_multi = True, False
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = 0, 0, 0
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "ri"
    _code.category = "Intronic"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeCapitalI:
    _code = _ClassCode("I")
    _code.definition = "Prediction completely contained within the introns of the reference transcript."
    _code.ref_multi, _code.pred_multi = True, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = 0, 0, 0
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "rI"
    _code.category = "Intronic"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeRI:
    _code = _ClassCode("ri")
    _code.definition = """Reverse intron transcript - the monoexonic reference is completely contained
    within one intron of the prediction transcript."""
    _code.ref_multi, _code.pred_multi = False, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = 0, 0, 0
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "i"
    _code.category = "Intronic"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeRCapitalI:
    _code = _ClassCode("rI")
    _code.definition = """Multiexonic reference completely contained within the introns of the prediction transcript."""
    _code.ref_multi, _code.pred_multi = True, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = 0, 0, 0
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "I"
    _code.category = "Intronic"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeF:
    _code = _ClassCode("f")
    _code.definition = """Fusion - this special code is applied when a prediction intersects more
    than one reference transcript. To be considered for fusions, candidate references must
    **either** share at least one splice junction with the prediction, **or** have at least 10% of
    its bases recalled. If two or more reference transcripts fit these constraints, then the
    prediction model is classified as a fusion."""
    _code.ref_multi, _code.pred_multi = None, None
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = "> 10", 0, 0
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = "> 0", 0, 0
    _code.reverse = None
    _code.category = "Fusion"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeX:
    _code = _ClassCode("x")
    _code.definition = "Monoexonic match on the **opposite** strand."
    _code.ref_multi, _code.pred_multi = None, False
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = ">0", ">0", ">0"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "x or X"
    _code.category = "Fragment"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeCapitalX:
    _code = _ClassCode("X")
    _code.definition = "Multiexonic match on the **opposite** strand."
    _code.ref_multi, _code.pred_multi = None, True
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = ">0", ">0", ">0"
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = None, None, None
    _code.reverse = "x or X"
    _code.category = "Fragment"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeP:
    _code = _ClassCode("p")
    _code.definition = """The prediction is on the same strand of a neighbouring but non-overlapping transcript.
    Probable polymerase run-on"""
    _code.ref_multi, _code.pred_multi = None, None
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = 0, 0, 0
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "p"
    _code.category = "Fragment"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeCapitalP:
    _code = _ClassCode("P")
    _code.definition = """The prediction is on the opposite strand of a neighbouring but non-overlapping transcript.
    Probable polymerase run-on."""
    _code.ref_multi, _code.pred_multi = None, None
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = 0, 0, 0
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = "P"
    _code.category = "Fragment"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


class CodeU:
    _code = _ClassCode("u")
    _code.definition = """Unknown - no suitable model has been found near enough the prediction to
    perform a comparison."""
    _code.ref_multi, _code.pred_multi = None, None
    _code._nucl_rec, _code._nucl_prec, _code._nucl_f1 = 0, 0, 0
    _code._junc_rec, _code._junc_prec, _code._junc_f1 = 0, 0, 0
    _code.reverse = None
    _code.category = "Unknown"

    code, definition = _code.code, _code.definition
    pred_multi, ref_multi = _code.pred_multi, _code.ref_multi
    category, reverse = _code.category, _code.reverse
    nucl, junc = _code.nucl, _code.junc

    __doc__ = str(_code)


codes = OrderedDict()
codes["="] = Equal
codes["_"] = UnderScore
codes["n"] = CodeN
codes["J"] = CodeCapitalJ
codes["c"] = CodeC
codes["C"] = CodeCapitalC
codes["j"] = CodeJ
codes["h"] = CodeH
codes["g"] = CodeG
codes["G"] = CodeCapitalG
codes["o"] = CodeO
codes["e"] = CodeE
codes["m"] = CodeM
codes["i"] = CodeI
codes["I"] = CodeCapitalI
codes["ri"] = CodeRI
codes["rI"] = CodeRCapitalI
codes["f"] = CodeF
codes["x"], codes["X"] = CodeX, CodeCapitalX
codes["p"], codes["P"], codes["u"] = CodeP, CodeCapitalP, CodeU

assert len(set(codes.values())) == len(codes), set.difference(set(codes.keys()),
                                                              set([_.code for _ in codes.values()]))
assert all(_ == codes[_].code for _ in codes)



from ...transcripts import Transcript
import msgpack
import zlib
from ..assignment.assigner import Assigner
import re
from queue import Queue
from logging import Logger
from typing import Union


def send_transcripts(transcript: Union[None,Transcript], rows: list, queue: Queue, maxsize=1000):
    if transcript is not None:
        transcript.finalize()
        d = transcript.as_dict(remove_attributes=False)
        rows.append(d)
        send_all = False
    else:
        send_all = True
    if len(rows) >= maxsize or send_all is True:
        to_write = msgpack.dumps(rows)
        queue.put(to_write)
        rows = []
    return rows


def get_best_result(transcript, assigner_instance: Assigner):
    transcript.finalize()
    assigner_instance.get_best(transcript)


orf_pattern = re.compile(r"\.orf[0-9]+$", re.IGNORECASE)


def transmit_transcript(transcript: Union[None,Transcript], done: int, lastdone: int,
                         rows: list, queue: Queue,
                         queue_logger: Logger,
                         __found_with_orf: set, send_all=False):

    if transcript is not None:
        if orf_pattern.search(transcript.id):
            __name = orf_pattern.sub("", transcript.id)
            if __name not in __found_with_orf:
                __found_with_orf.add(__name)
                done += 1
                if done and done % 10000 == 0:
                    queue_logger.info("Parsed %s transcripts", done)
                    lastdone = done
                rows = send_transcripts(transcript, rows, queue)
            else:
                pass
        else:
            done += 1
            if done and done % 10000 == 0:
                queue_logger.info("Parsed %s transcripts", done)
                lastdone = done
            try:
                rows = send_transcripts(transcript, rows, queue)
            except AssertionError:
                raise AssertionError((transcript.id, ))

    if send_all is True:
        send_transcripts(None, rows, queue)

    return rows, done, lastdone, __found_with_orf



# coding: utf-8

"""
This class defines the results of the Assigner.compare method.
"""


import itertools


def _convert_item(item, final_type):
    it_type = type(item)
    if it_type == bytes:
        item = it_type.decode()
        it_type = str
    if item == "-":  # Null value.
        return [item]
    elif it_type == str and "," in item:
        return list(itertools.chain.from_iterable([_convert_item(_, final_type) for _ in item.split(",")]))
    else:
        if final_type is not None:
            return [final_type(item)]
        else:
            return [item]


def _tupling(obj, final_type):
    tup_type = type(obj)
    if tup_type in (tuple, list):
        return tuple(itertools.chain.from_iterable([_convert_item(item, final_type) for item in obj]))
    elif tup_type in (str, float, int, bytes):
        return tuple(_convert_item(obj, final_type))
    else:
        raise TypeError(type(obj))


def _to_float_tuple(tup):
    return _tupling(tup, float)


def _to_int_tuple(tup):
    return _tupling(tup, int)


def _to_tuple(tup):
    return _tupling(tup, None)


class ResultStorer:
    """This class stores the results in pre-defined slots, to reduce memory usage."""

    __name__ = "resultstorer"

    __slots__ = ["ref_id", "ref_gene", "ccode",
                 "tid", "gid",
                 "tid_num_exons", "ref_num_exons",
                 "n_prec", "n_recall", "n_f1",
                 "j_prec", "j_recall", "j_f1",
                 "e_prec", "e_recall", "e_f1",
                 "distance",
                 "location"]

    @staticmethod
    def to_float_tuple(tup):
        return _to_float_tuple(tup)

    @staticmethod
    def to_int_tuple(tup):
        return _to_int_tuple(tup)

    @staticmethod
    def to_tuple(tup):
        return _to_tuple(tup)

    _types = [_to_tuple, _to_tuple, _to_tuple,
              str, str, int, _to_int_tuple,
              _to_float_tuple, _to_float_tuple, _to_float_tuple,
              _to_float_tuple, _to_float_tuple, _to_float_tuple,
              _to_float_tuple, _to_float_tuple, _to_float_tuple,
              _to_int_tuple, str]

    @property
    def types(self):
        return self._types

    def __init__(self, *args, state=None):

        """
        :param args: a list/tuple
        :type args: list | tuple | str | int | float

        """

        if state is not None and len(args) == 0:
            self._load_dict(state)
            return
        
        if len(args) != len(self.__slots__):
            err_msg = "Result_storer expected {0} but only received {1}".format(
                len(self.__slots__), len(args))
            raise ValueError(err_msg)

        self.ref_id, self.ref_gene, self.ccode, self.tid, self.gid, \
            self.tid_num_exons, self.ref_num_exons, \
            self.n_prec, self.n_recall, self.n_f1,\
            self.j_prec, self.j_recall, self.j_f1, \
            self.e_prec, self.e_recall, self.e_f1, \
            self.distance, self.location = args

        for index, key in enumerate(self.__slots__):
            setattr(self, key, self.types[index](getattr(self, self.__slots__[index])))

    def _load_dict(self, state):

        for index, key in enumerate(self.__slots__):
            setattr(self, key, self.types[index](state[key]))
        return
                    
    def _asdict(self):

        """
        :return: a dictionary containing the items of the class
        :rtype : dict
        """
        result_dict = dict().fromkeys(self.__slots__)
        for attr in self.__slots__:
            result_dict[attr] = getattr(self, attr)
            if isinstance(result_dict[attr], (tuple, list)):
                result_dict[attr] = ",".join(["{}".format(_) for _ in result_dict[attr]])
        return result_dict

    def as_dict(self):
        """
        Wrapper for the protected method _asdict
        :return: dictionary
        """

        return self._asdict()

    def __str__(self):

        result_dict = self._asdict()
        line = []
        for key in self.__slots__:
            line.append(str(result_dict[key]))
        return "\t".join(line)

    def __repr__(self):

        represent = "result( "
        for key in self.__slots__:
            represent += "{0}={1}, ".format(key, getattr(self, key))
        represent += ")"
        return represent

    def __getitem__(self, item):
        if item in self.__slots__:
            return getattr(self, item)
        else:
            raise KeyError(item)



#!/usr/bin/env python3
# coding: utf-8

"""Script to trim down the terminal exons of multiexonic transcripts"""

import argparse
import sys
from ...utilities.log_utils import create_default_logger
from ...exceptions import InvalidTranscript


__author__ = 'Luca Venturini'


def trim_noncoding(transcript, max_length=0):
    """
    Function to trim the terminal exons for non-coding transcripts,
     i.e. the simplest case.

    :param transcript: the transcript to be trimmed.
    :type transcript: Transcript

    :param max_length: maximum length of terminal exons.
    :type max_length: int

    :return: transcript
    :rtype: Transcript

    """

    # Non-coding transcript: trim the terminal exons and the return

    transcript.unfinalize()
    exons = sorted(transcript.exons)
    first, last = exons[0], exons[-1]
    if (first[1] - first[0] + 1) > max_length:
        newfirst = tuple([first[1] - max_length, first[1]])
        transcript.start = newfirst[0]
        transcript.remove_exon(first)
        transcript.add_exon(newfirst)

    if (last[1] - last[0] + 1) > max_length:
        newlast = tuple([last[0], last[0] + max_length])
        transcript.end = last[0] + max_length
        transcript.remove_exon(last)
        transcript.add_exon(newlast)

    return transcript


def trim_start(transcript, cds_start, max_length=0):

    """
    Method to trim the beginning of a transcript, given both
    the cds_start and the maximal allowed length after trimming.
    :param transcript: the transcript to be trimmed
    :type transcript: Transcript

    :param cds_start: the position at which the CDS starts for this transcript.
    :type cds_start: int

    :param max_length: maximum length of terminal exons.
    :type max_length: int

    :rtype: Transcript

    """

    first = transcript.exons[0]
    if first[1] < cds_start:  # first exon is only UTR
        if first[1] - first[0] + 1 > max_length:
            # First exon longer than max_length; trim it,
            # remove from UTR and exons, substitute with trimmed exon
            newfirst = tuple([first[1] - max_length, first[1]])
            transcript.remove_exon(first)
            transcript.add_exon(newfirst)
        else:
            # Leave as it is
            newfirst = first
    elif first[0] < cds_start <= first[1]:
        # Partly UTR partly CDS
        if first[1] - first[0] + 1 > max_length:
            newfirst = tuple([min(cds_start, first[1] - max_length), first[1]])
            # Retrieve and remove offending UTR segment
            transcript.remove_exon(first)
            transcript.add_exon(newfirst)
            transcript.add_exon((cds_start, newfirst[1]), feature="CDS")
        else:
            newfirst = first
    else:  # Beginning of first exon == beginning of CDS
        newfirst = first
    transcript.start = newfirst[0]
    return transcript


def trim_end(transcript, cds_end, max_length=0):
    """
    Method to trim the end of a transcript, given both
    the cds_end and the maximal allowed length after trimming.
    :param transcript: the transcript to analyse
    :type transcript: Transcript

    :param cds_end: position of CDS end
    :type cds_end: int

    :param max_length: maximum length of terminal exons.
    :type max_length: int

    :return: transcript
    :rtype: Mikado.loci_objects.transcript.Transcript
    """

    # We have to sort b/c we appended a new exon in trim_start
    last = sorted(transcript.exons)[-1]
    if last[0] > cds_end:  # Only UTR
        if last[1] - last[0] + 1 > max_length:

            newlast = tuple([last[0], last[0] + max_length])
            if last != newlast:
                transcript.remove_exon(last)
                transcript.add_exon(newlast)
        else:
            newlast = last
    elif last[0] <= cds_end < last[1]:
        if last[1] - last[0] + 1 > max_length:
            newlast = tuple([last[0], max(cds_end, last[0] + max_length)])
            transcript.remove_exon(last)
            transcript.add_exon(newlast)
            transcript.add_exon((newlast[0], cds_end), feature="CDS")
        else:
            newlast = last
    else:
        newlast = last

    transcript.end = newlast[1]
    assert all([(exon[1] <= newlast[1] for exon in transcript.exons)])
    return transcript


def trim_coding(transcript, logger, max_length=0):
    """
    Function to trim the terminal exons for coding transcripts,
     i.e. the more complex case as we have to account for CDS
     as well.
    :param transcript: the transcript to analyse.
    :type transcript: Transcript

    :param logger: a logger instance
    :type logger: logging.Logger

    :param max_length: maximum length of terminal exons.
    :type max_length: int

    :return: transcript
    :rtype: Transcript
    """
    # Coding transcript
    # Order cds_start and end irrespectively of strand

    # Wrong CDS

    if transcript.selected_cds_start is None:
        logger.debug("Non-coding transcript %s submitted, calling the correct function.",
                     transcript.id)
        return trim_noncoding(transcript, max_length=max_length)

    cds_start, cds_end = sorted([transcript.selected_cds_end,
                                 transcript.selected_cds_start])

    transcript.unfinalize()
    assert transcript.internal_orfs == []
    if cds_start == cds_end:
        logger.warning("{0} has coincident CDS start and end coordinates. Stripping CDS".format(
            transcript.id))
        transcript.strip_cds()
        transcript.feature = "transcript"
        # Now we have to re-deinitialise the transcript
        transcript.finalized = False
        transcript = trim_noncoding(transcript,
                                    max_length=max_length)
        return transcript

    assert cds_start < cds_end, "\n".join([str(_) for _ in (cds_start, cds_end, transcript)])
    transcript = trim_start(transcript, cds_start,
                            max_length=max_length)
    transcript = trim_end(transcript, cds_end,
                          max_length=max_length)

    return transcript


def strip_terminal(transcript, args):
    """This function will take as input a transcript and then:
    - return immediately if the transcript is monoexonic
    - trim the terminal exons to a length of max(args.max_length (if longer), CDS boundary)

    :param transcript: the input transcript
    :type transcript: Transcript

    :param args: argparse configuration namespace
    :type args: argparse.Namespace

    :rtype : Transcript
    """

    transcript.finalize()
    # Return immediately if the transcript is monoexonic
    if transcript.monoexonic is True:
        return transcript

    if transcript.selected_cds_start is not None:
        coding = True
    else:
        coding = False

    # noinspection PyUnresolvedReferences
    max_l = args.max_length
    try:
        if coding:
            assert sorted([transcript.selected_cds_start,
                           transcript.selected_cds_end])
            transcript = trim_coding(transcript, args.logger, max_length=max_l)
        else:
            transcript = trim_noncoding(transcript, max_length=max_l)
    except InvalidTranscript as exc:
        args.logger.exception(exc)
        return None

    # noinspection PyUnresolvedReferences
    transcript.finalize()
    assert transcript.start == transcript.exons[0][0], (transcript.start, transcript.exons)
    assert transcript.end == transcript.exons[-1][1], (transcript.end, transcript.exons)
    return transcript


def trim_gene(current_gene, format_name, args):
    """"""

    if current_gene is not None:
        current_gene.finalize()
        new_start, new_end = float("inf"), float("-inf")
        for tid in list(current_gene.transcripts)[:]:
            trimmed = strip_terminal(current_gene[tid], args)
            current_gene.remove(tid)  # We have to remove the old transcript anyway
            if trimmed is not None:
                current_gene.add(trimmed)
                new_start, new_end = min(new_start,
                                         trimmed.start), max(new_end,
                                                             trimmed.end)
        # Reset them manually to avoid warnings
        current_gene.start, current_gene.end = new_start, new_end
        current_gene.finalize()
        if len(current_gene.transcripts) > 0:
            print(current_gene.format(format_name), file=args.out)
    return


def launch(args):

    """
    Launcher of the program.
    :param args: the argparse Namespace.
    """

    from ...parsers import parser_factory
    from ...transcripts import Transcript, Gene

    args.gff = parser_factory(args.gff)
    args.logger = create_default_logger("trimmer")
    args.logger.setLevel("WARN")

    if args.as_gtf is False:
        print("##gff-version 3", file=args.out)

    if args.as_gtf is True:
        format_name = "gtf"
    else:
        format_name = "gff3"

    if args.ann.file_format == "gff3":
        current_gene = None
        for record in args.ann:
            if record.is_gene is True:
                trim_gene(current_gene, format_name, args)
                current_gene = Gene(record, logger=args.logger)
            elif record.is_transcript is True:
                assert current_gene is not None
                assert record.parent[0] == current_gene.id
                current_gene.add(Transcript(record))
            elif record.is_exon is True:
                current_gene.add_exon(record)
            elif record.is_derived is True:
                for derived in record.derived_from:
                    current_gene[derived].add_derived_child(record.id)

        trim_gene(current_gene, format_name, args)
    else:
        # When analysing a GTF, we can blissfully ignore everything
        current_gene = None
        for record in args.ann:
            if current_gene is not None and current_gene.id == record.gene:
                if record.is_exon is True:
                    if record.transcript in current_gene:
                        current_gene.add_exon(record)
                    else:
                        current_gene.add(Transcript(record))
                elif record.is_transcript is True:
                    assert record.transcript not in current_gene
                    current_gene.add(Transcript(record))
            else:
                trim_gene(current_gene, format_name, args)
                current_gene = Gene(Transcript(record), logger=args.logger)
            continue
        trim_gene(current_gene, format_name, args)


def trim_parser():

    """
    Command line parser for the utility.
    """

    parser = argparse.ArgumentParser("Script to trim down the terminal exons of multiexonic transcripts")
    parser.add_argument("-ml", "--max_length", type=int, default=50,
                        help="Maximal length of trimmed terminal exons")
    parser.add_argument("--as-gtf", default=False, action="store_true",
                        dest="as_gtf",
                        help="Flag. If set, the output will be in GTF rather than GFF3 format.")
    parser.add_argument("ann", type=str, help="Reference GTF/GFF output file.")
    parser.add_argument("out", nargs="?", default=sys.stdout,
                        type=argparse.FileType('w'))
    parser.set_defaults(func=launch)
    return parser



import zlib
from time import perf_counter
from multiprocessing import Process
from typing import Union
from multiprocessing.managers import AutoProxy
import logging
from itertools import product
import logging.handlers as logging_handlers
import functools
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, sessionmaker
from ..configuration import MikadoConfiguration, DaijinConfiguration
from ..utilities import dbutils
from ..scales.assignment.assigner import Assigner
from ..loci.superlocus import Superlocus
from ._merge_loci_utils import __create_gene_counters, manage_index
import collections
import sys
from ..transcripts import Transcript
from ..exceptions import InvalidTranscript, InvalidConfiguration, InvalidCDS, NotInLocusError
from ..parsers.GTF import GtfLine
from ..configuration.configurator import load_and_validate_config
import msgpack
from ._loci_serialiser import serialise_locus


__author__ = 'Luca Venturini'


def merge_loci(mapper,
               out_handles,
               total,
               configuration,
               logger=None,
               source="Mikado"):

    """ Function to merge the temporary loci files into single output files,
      renaming the genes according to the preferred style.
    :param num_temp: number of temporary files.
    :param out_handles: The names of the output loci files.
    :param logger: logger to use
    :return:
    """

    locus_metrics, locus_scores, locus_out = out_handles[0]
    sub_metrics, sub_scores, sub_out = out_handles[1]
    mono_metrics, mono_scores, mono_out = out_handles[2]

    if len(mapper["results"]) != total:
        raise KeyError("I am missing some loci! {} vs {}".format(len(mapper["results"]), total))

    # Start iterating the output dictionaries ("cursors")
    print_subloci = (out_handles[1][0] is not None)
    print_monoloci = (out_handles[2][0] is not None)
    __valid = set(range(1, total + 1))

    if set(mapper["results"].keys()) != __valid:
        missing = set.difference(__valid, set(mapper["results"].keys()))
        raise AssertionError("Missing the following loci: {}".format(missing))

    new_common, total_genes = __create_gene_counters(mapper["results"])

    done = 0
    logger.info("We have a total of %d genes", total_genes)
    tot_loci = set()
    for index in sorted(mapper["results"]):
        loci, batch = manage_index((index, new_common[index]), mapper["results"], source=source,
                                   configuration=configuration)
        done += 1
        if loci and set(loci).issubset(tot_loci):
            raise ValueError("Duplicated loci! {}".format(loci))

        tot_loci.update(set(loci))
        for minibatch in batch:
            if minibatch[0] is not None:
                locus_lines, locus_metrics_rows, locus_scores_rows = minibatch[0]
                if locus_lines:
                    assert len(locus_metrics_rows) > 0
                    print(locus_lines, file=locus_out)
                for row in locus_metrics_rows:
                    print(*[row[key] for key in locus_metrics.fieldnames],
                          sep="\t", file=locus_metrics.handle)
                for row in locus_scores_rows:
                    print(*[row[key] for key in locus_scores.fieldnames],
                          sep="\t", file=locus_scores.handle)

            if print_subloci and minibatch[1]:
                sub_lines, sub_metrics_rows, sub_scores_rows = minibatch[1]
                if sub_lines != '':
                    print(sub_lines, file=sub_out)

                for row in sub_metrics_rows:
                    try:
                        print(*[row[key] for key in sub_metrics.fieldnames],
                              sep="\t", file=sub_metrics.handle)
                    except KeyError:
                        raise KeyError("\n".join([str((key, str(row.get(key, "MISSING"))))
                                                  for key in sub_metrics.fieldnames]))
                for row in sub_scores_rows:
                    print(*[row[key] for key in sub_scores.fieldnames],
                              sep="\t", file=sub_scores.handle)
            if print_monoloci and minibatch[2]:
                mono_lines, mono_metrics_rows, mono_scores_rows = minibatch[2]
                if mono_lines != '':
                    print(mono_lines, file=mono_out)
                for row in mono_metrics_rows:
                    print(*[row[key] for key in mono_metrics.fieldnames],
                          sep="\t", file=mono_metrics.handle)
                for row in mono_scores_rows:
                    print(*[row[key] for key in mono_scores.fieldnames],
                          sep="\t", file=mono_scores.handle)

    if done != total:
        raise KeyError("Something has been lost")

    return


def remove_fragments(stranded_loci, configuration, logger):

    """This method checks which loci are possible fragments, according to the
    parameters provided in the configuration file, and tags/remove them according
    to the configuration specification.

    :param stranded_loci: a list of the loci to consider for fragment removal
    :type stranded_loci: list[Superlocus]

    :param configuration: the configuration dictionary
    :type configuration: (MikadoConfiguration|DaijinConfiguration)

    :param logger: the logger
    :type logger: logging.Logger

    """

    loci_to_check = {True: list(), False: list()}

    total = 0

    stranded_loci_dict = dict()
    loci_to_superloci = dict()

    start, end = float("inf"), float("-inf")
    chrom = None

    for stranded_locus in stranded_loci:
        stranded_loci_dict[stranded_locus.id] = stranded_locus
        start, end = min(stranded_locus.start, start), max(stranded_locus.end, end)
        chrom = chrom or stranded_locus.chrom
        for _, locus_instance in stranded_locus.loci.items():
            loci_to_superloci[locus_instance.id] = stranded_locus.id
            logger.debug("Assessing whether %s could be a fragment", _)
            total += 1
            is_fragment = locus_instance.is_putative_fragment()
            logger.debug("%s is a putative fragment: %s", _, is_fragment)
            loci_to_check[is_fragment].append(locus_instance)

    if len(loci_to_check[True]) == total:
        loci_to_check[False] = loci_to_check.pop(True)
        loci_to_check[True] = list()

    comparisons = collections.defaultdict(list)
    # Produce a list of duples

    for locus_to_check, gene in product(loci_to_check[True], loci_to_check[False]):
        is_to_be_filtered, comparison = gene.other_is_fragment(locus_to_check)
        if is_to_be_filtered is True:
            comparisons[locus_to_check.id].append(comparison)

    for locus in comparisons:
        if configuration.pick.fragments.remove is True:
            # A bit convoluted: use the locus ID to find the correct superlocus, then delete the ID inside the SL.
            if locus not in stranded_loci_dict[loci_to_superloci[locus]].loci:
                logger.error("Locus %s has been lost from superlocus %s!",
                             locus,
                             loci_to_superloci[locus])
                continue
            del stranded_loci_dict[loci_to_superloci[locus]].loci[locus]
            # stranded_loci_dict[loci_to_superloci[locus]].loci.pop(locus, None)
        else:
            best_comparison = sorted(comparisons[locus], reverse=True, key=Assigner.get_f1)[0]
            stranded_loci_dict[loci_to_superloci[locus]].loci[locus].is_fragment = True
            stranded_loci_dict[loci_to_superloci[locus]].loci[locus].attributes[
                "fragment_of"] = best_comparison.ref_id[0]
            stranded_loci_dict[loci_to_superloci[locus]].loci[locus].attributes[
                "fragment_class_code"] = best_comparison.ccode[0]
            if best_comparison.distance[0] > 0:
                stranded_loci_dict[loci_to_superloci[locus]].loci[locus].attributes[
                    "distance"] = best_comparison.distance[0]

    for stranded_locus in stranded_loci:
        yield stranded_locus


def analyse_locus(slocus: Superlocus,
                  counter: int,
                  configuration: Union[MikadoConfiguration,DaijinConfiguration],
                  logging_queue: AutoProxy,
                  engine=None,
                  session=None) -> [Superlocus]:

    """
    :param slocus: a superlocus instance
    :type slocus: Mikado.loci_objects.superlocus.Superlocus

    :param counter: an integer which is used to create the proper name for the locus.
    :type counter: int

    :param configuration: the configuration dictionary
    :type configuration: (MikadoConfiguration|DaijinConfiguration)

    :param logging_queue: the logging queue
    :type logging_queue: multiprocessing.managers.AutoProxy

    :param engine: an optional engine to connect to the database.
    :type data_dict: sqlalchemy.engine.engine

    This function takes as input a "superlocus" instance and the pipeline configuration.
    It also accepts as optional keywords a dictionary with the CDS information
    (derived from a Bed12Parser) and a "lock" used for avoiding writing collisions
    during multithreading.
    The function splits the superlocus into its strand components and calls the relevant methods
    to define the loci.
    When it is finished, it transmits the superloci to the printer function.
    """

    # Define the logger
    if slocus is None:
        # printer_dict[counter] = []
        return []

    handler = logging_handlers.QueueHandler(logging_queue)
    logger = logging.getLogger("{0}:{1}-{2}".format(
        slocus.chrom, slocus.start, slocus.end))
    logger.addHandler(handler)

    # We need to set this to the lowest possible level,
    # otherwise we overwrite the global configuration
    logger.setLevel(configuration.log_settings.log_level)
    logger.propagate = False
    logger.debug("Started with %s, counter %d", slocus.id, counter)
    if slocus.stranded is True:
        logger.warning("%s is stranded already! Resetting", slocus.id)
        slocus.stranded = False

    slocus.logger = logger
    slocus.source = configuration.pick.output_format.source
    logger.info("Starting to process %s", slocus.id)
    start_time = perf_counter()
    slocus.load_all_transcript_data(engine=engine, session=session)
    logger.debug("Loading transcript data for %s", slocus.id)

    # Load the CDS information if necessary
    if slocus.initialized is False:
        # This happens when all transcripts have been removed from the locus,
        # due to errors that have been hopefully logged
        logger.warning(
            "%s had all transcripts failing checks, ignoring it",
            slocus.id)
        return []

    # Split the superlocus in the stranded components
    logger.debug("Splitting by strand")
    stranded_loci = sorted([_ for _ in slocus.split_strands()])
    # Define the loci
    logger.debug("Divided into %d loci", len(stranded_loci))

    failed = 0
    original_size = len(stranded_loci)
    for stranded_locus in stranded_loci:
        stranded_locus.logger = logger
        try:
            stranded_locus.define_loci()
        except (InvalidTranscript, InvalidConfiguration, InvalidCDS, NotInLocusError) as exc:
            logger.exception(exc)
            logger.error("Removing failed locus %s", stranded_locus.name)
            stranded_loci.remove(stranded_locus)
            assert stranded_locus not in stranded_loci
            failed += 1
        logger.debug("Defined loci for %s:%f-%f, strand: %s",
                     stranded_locus.chrom,
                     stranded_locus.start,
                     stranded_locus.end,
                     stranded_locus.strand)
        assert all(locus.metrics_calculated is True for locus in stranded_locus.loci.values()), \
            "Metrics not calculated after analysis in analyse_locus."
        assert all(locus.scores_calculated is True for locus in stranded_locus.loci.values()), \
            "Scores not calculated after analysis in analyse_locus."

    assert len(stranded_loci) + failed == original_size, (len(stranded_loci), failed, original_size)
    # Check if any locus is a fragment, if so, tag/remove it
    if len(stranded_loci) > 0:
        stranded_loci = sorted(list(remove_fragments(stranded_loci, configuration, logger)))
    logger.debug("Size of the loci to send: {0}, for {1} loci".format(sys.getsizeof(stranded_loci),
                                                                      len(stranded_loci)))
    end_time = perf_counter()
    logger.info("Finished with locus %s, counter %d in %f seconds", slocus.id, counter, end_time - start_time)
    logger.removeHandler(handler)
    handler.close()
    return stranded_loci


class LociProcesser(Process):

    """This process class takes care of getting from the queue the loci,
    analyse them, and print them to the output files."""

    def __init__(self,
                 configuration,
                 locus_queue,
                 logging_queue,
                 status_queue,
                 identifier,
                 ):

        super(LociProcesser, self).__init__()
        configuration = load_and_validate_config(msgpack.loads(configuration, raw=False))
        self.logging_queue = logging_queue
        self.status_queue = status_queue
        self.__identifier = identifier  # Property directly unsettable
        self.name = "LociProcesser-{0}".format(self.identifier)
        self.configuration = configuration
        for section in (self.configuration.scoring.requirements, self.configuration.scoring.as_requirements,
                        self.configuration.scoring.cds_requirements, self.configuration.scoring.not_fragmentary):
            # Compile the expression
            _ = section.compiled

        self.engine = None
        self.session = None
        self.handler = logging_handlers.QueueHandler(self.logging_queue)
        self.logger = logging.getLogger(self.name)
        self.logger.addHandler(self.handler)
        self.logger.setLevel(self.configuration.log_settings.log_level)

        self.logger.propagate = False
        self.locus_queue = locus_queue
        self.logger.debug("Starting Process %s", self.name)

        self.logger.debug("Starting the pool for {0}".format(self.name))

    @property
    def identifier(self):
        return self.__identifier

    def __getstate__(self):

        state = self.__dict__.copy()
        state["engine"] = None
        state["session"] = None
        state["analyse_locus"] = None
        state["logger"].removeHandler(self.handler)
        del state["handler"]
        if "logger" in state:
            del state["logger"]
        return state

    def terminate(self):
        self.__close_handles()
        super().terminate()

    def kill(self) -> None:
        self.__close_handles()
        super(LociProcesser, self).kill()

    def __close_handles(self):
        """Private method to flush and close all handles."""
        if self.engine is not None:
            self.engine.dispose()
        if self.session is not None:
            self.session.close_all()

    def close(self):
        self.__close_handles()
        self.join()

    def __setstate__(self, state):

        self.__dict__.update(state)
        self.handler = logging_handlers.QueueHandler(self.logging_queue)
        self.logger = logging.getLogger(self.name)
        self.logger.addHandler(self.handler)
        self.logger.setLevel(self.configuration.log_settings.log_level)
        self.logger.propagate = False
        self.engine = dbutils.connect(self.configuration, self.logger)
        self.handler = logging_handlers.QueueHandler(self.logging_queue)
        self.logger.addHandler(self.handler)
        self.connect_to_db(engine=None, session=None)

    def join(self, timeout=None):
        self.__close_handles()
        # self.terminate()
        super().join(timeout=timeout)

    def connect_to_db(self, engine: Union[Engine, None], session: Union[Session, None]):

        """
        :param engine: the connection pool
        :type engine: Engine

        :param session: a preformed session
        :type session: Session

        This method will connect to the database using the information
        contained in the JSON configuration.
        """

        if isinstance(session, Session):
            self.session = session
            self.sessionmaker = sessionmaker()
            self.sessionmaker.configure(bind=self.session.bind)
            self.engine = self.session.bind

        if engine is None:
            self.engine = dbutils.connect(self.configuration)
        else:
            self.engine = engine

        self.sessionmaker = sessionmaker()
        self.sessionmaker.configure(bind=self.engine)
        self.session = self.sessionmaker()

    def run(self):
        """Start polling the queue, analyse the loci, and send them to the printer process."""
        self.connect_to_db(None, None)
        self.logger.debug("Starting to parse data for {0}".format(self.name))
        self.analyse_locus = functools.partial(analyse_locus,
                                               configuration=self.configuration,
                                               engine=self.engine,
                                               session=self.session,
                                               logging_queue=self.logging_queue)

        print_cds = (not self.configuration.pick.run_options.exclude_cds)

        print_monoloci = (self.configuration.pick.files.monoloci_out is not None and
                          len(self.configuration.pick.files.monoloci_out) > 0)
        print_subloci = (self.configuration.pick.files.subloci_out is not None and
                         len(self.configuration.pick.files.subloci_out) > 0)

        accumulator = []
        while True:
            vals = self.locus_queue.get()
            try:
                counter, transcripts = vals
            except ValueError:
                raise ValueError(vals)

            if counter == "EXIT":
                self.logger.debug("EXIT received for %s", self.name)
                accumulator = zlib.compress(msgpack.dumps(accumulator))
                self.status_queue.put(accumulator)
                self.logger.debug("Put all the remaining results into the queue for %s (size %s)", self.name,
                                  sys.getsizeof(accumulator))
                # self.locus_queue.task_done()
                self.locus_queue.put((counter, None))
                accumulator = []
                break

            try:
                try:
                    transcripts = msgpack.loads(transcripts, raw=False)
                except TypeError as err:
                    raise TypeError("{}, {}".format(err, transcripts))
                if len(transcripts) == 0:
                    stranded_loci = []
                    self.logger.warning("No transcript found for index %d", counter)
                else:
                    tobjects = []
                    chroms = set()
                    for tjson in transcripts:
                        definition = GtfLine(tjson["definition"]).as_dict()
                        is_reference = definition["source"] in self.configuration.prepare.files.reference
                        transcript = Transcript(logger=self.logger,
                                                source=definition["source"],
                                                intron_range=self.configuration.pick.run_options.intron_range,
                                                is_reference=is_reference)
                        transcript.chrom, transcript.start, transcript.end = (definition["chrom"],
                                                                              definition["start"], definition["end"])
                        chroms.add(transcript.chrom)
                        assert len(chroms) == 1, chroms
                        transcript.id = definition["transcript"]
                        transcript.strand, transcript.feature = definition["strand"], definition["feature"]
                        transcript.attributes = definition["attributes"]
                        try:
                            for exon in tjson["exon_lines"]:
                                start, end, feature, phase = exon
                                transcript.add_exon((start, end), feature=feature, phase=phase)
                            transcript.finalize()
                            tobjects.append(transcript)
                        except InvalidTranscript as exc:
                            self.logger.exception("Transcript %s is invalid. Ignoring. Error: %s",
                                                  transcript.id, exc)

                    slocus = Superlocus(tobjects.pop(),
                                        stranded=False,
                                        configuration=self.configuration,
                                        source=self.configuration.pick.output_format.source)
                    slocus.add_transcripts_to_locus(tobjects, check_in_locus=False)
                    stranded_loci = self.analyse_locus(slocus, counter)

                if len(stranded_loci) == 0:
                    self.logger.warning("No loci left for index %d", counter)
                    accumulator = serialise_locus([], accumulator, counter, print_cds=print_cds,
                                                  print_monosubloci=print_monoloci, print_subloci=print_subloci)
                else:
                    if len(accumulator) >= 100:
                        accumulator = zlib.compress(msgpack.dumps(accumulator))
                        self.status_queue.put(accumulator)
                        accumulator = []

                    accumulator = serialise_locus(
                        stranded_loci, accumulator, counter, print_cds=print_cds,
                        print_monosubloci=print_monoloci, print_subloci=print_subloci)
            except (InvalidTranscript, InvalidCDS, AssertionError, ValueError, RuntimeError, TypeError) as exc:
                self.logger.error(f"{self.name} has encountered an error, details:\n{exc}")
                current = len(accumulator)
                accumulator = serialise_locus([], accumulator, counter, print_cds=print_cds,
                                              print_monosubloci=print_monoloci, print_subloci=print_subloci)
                assert len(accumulator) > current
            finally:
                self.locus_queue.task_done()

        if accumulator:
            assert isinstance(accumulator, list)
            self.logger.error(accumulator)
            accumulator = zlib.compress(msgpack.dumps(accumulator))
            self.status_queue.put(accumulator)
        self.logger.info(f"{self.name} has finished")
        return



"""
This module implements the HSP serialisation class.
"""

from sqlalchemy import Column, String, Integer, Float, ForeignKey, Index
from sqlalchemy.sql.schema import PrimaryKeyConstraint, UniqueConstraint
from sqlalchemy.orm import relationship, column_property, backref
from sqlalchemy import select
from sqlalchemy.ext.hybrid import hybrid_property  # hybrid_method
from ...utilities.dbutils import DBBASE
from .query import Query
from .target import Target
from .aln_string_parser import prepare_aln_strings


__author__ = 'Luca Venturini'


class Hsp(DBBASE):

    r"""
    This class serializes and stores into the DB the various HSPs.
    It is directly connected to the Hit table, through the "hit_id"
    reference key.
    The Hit reference can be accessed through the hit_object attribute;
    back-reference (Hit to Hsps) is given by the "hsps" attribute.

    Keys:

    :return hit_id: Reference key for the Hit table
    :rtype hit_id: int

    :return counter: It indicates the progressive number of the HSP for the hit
    :rtype counter: int

    :return query_hsp_start: Start position on the query
    :rtype query_hsp_start; int

    :return query_hsp_end: End position on the query
    :rtype query_hsp_end: int

    :return target_hsp_start: Start position on the target
    :rtype target_hsp_start: int

    :return target_hsp_end: End position on the target
    :rtype target_hsp_end: int

    :return hsp_evalue: Evalue of the HSP
    :rtype hsp_evalue: float

    :return hsp_bits: Bit-score of the HSP
    :rtype hsp_bits: float

    :return hsp_identity: Identity (in %) of the alignment
    :rtype  hsp_identity: float

    :return hsp_length: Length of the HSP
    :rtype hsp_length: int

    :return match: the match line between query and target, with the following specs:
        - If the position is a match/positive, keep the original value
        - If the position is a gap *for the query*, insert a - (dash)
        - If the position is a gap *for the target*, insert a _ (underscore)
        - If the position is a gap *for both*, insert a \ (backslash)

    An HSP row has the following constraints:
    - Counter,hit_id must be unique (and are primary keys)
    - The combination ("Hit_id","query_hsp_start","query_hsp_end",
    "target_hsp_start", "target_hsp_end") must be unique

    Moreover, the following properties are also present:

    :return query_object: The referenced Query
    :rtype query_object: Query

    :return target_object: The reference Target
    :rtype target_object: Target
    """

    __tablename__ = "hsp"
    counter = Column(Integer)  # Indicates the number of the HSP inside the hit
    query_id = Column(Integer, ForeignKey(Query.query_id), unique=False)
    target_id = Column(Integer, ForeignKey(Target.target_id), unique=False)
    query_hsp_start = Column(Integer)
    query_hsp_end = Column(Integer)
    query_frame = Column(Integer)
    target_hsp_start = Column(Integer)
    target_hsp_end = Column(Integer)
    target_frame = Column(Integer)
    match = Column(String(100000))
    hsp_evalue = Column(Float)
    hsp_bits = Column(Float)
    hsp_identity = Column(Float)
    hsp_positives = Column(Float)
    hsp_length = Column(Integer)

    # Indices and constraints
    pk_constraint = PrimaryKeyConstraint("counter",
                                         "query_id",
                                         "target_id",
                                         name="hsp_constraint")
    query_index = Index("hsp_query_idx", "query_id", unique=False)
    target_index = Index("hsp_target_idx", "target_id", unique=False)
    hsp_evalue_index = Index('hsp_evalue_idx', 'hsp_evalue', unique=False)
    combined_index = Index("hsp_combined_idx",
                           "query_id",
                           "target_id", unique=False)
    full_index = Index("hsp_full_idx",
                       "counter", "query_id", "target_id", unique=True)
    uni_constraint = UniqueConstraint("query_id", "target_id",
                                      "query_hsp_start", "query_hsp_end",
                                      "target_hsp_start", "target_hsp_end")

    __table_args__ = (pk_constraint, query_index, target_index, combined_index, hsp_evalue_index, uni_constraint)

    query_object = relationship(Query, uselist=False,
                                lazy="select",
                                innerjoin=True,
                                backref=backref("hsps", cascade="all, delete-orphan"))
    target_object = relationship(Target,
                                 uselist=False,
                                 lazy="select",
                                 innerjoin=True,
                                 backref=backref("hsps", cascade="all, delete-orphan"))

    def __init__(self, counter, query_id, target_id, query_hsp_start, query_hsp_end, query_frame, target_hsp_start,
                 target_hsp_end, target_frame,
                 match, hsp_evalue, hsp_bits, hsp_identity, hsp_positives, hsp_length):

        """

        :param hsp: an hsp object from the serialized XML

        :param counter: integer which indicates the HSP position in the HSP
         list for the hit
        :type counter: int

        :param query_id: Foreign key for the Query table
        :type query_id: int

        :param target_id: Foreign key for the Target table
        :type target_id: int
        """

        self.counter = counter
        self.query_id = query_id
        self.target_id = target_id
        self.query_hsp_start, self.query_hsp_end, self.query_frame = query_hsp_start, query_hsp_end, query_frame
        self.target_hsp_start, self.target_hsp_end, self.target_frame = target_hsp_start, target_hsp_end, target_frame
        self.match, self.hsp_length = match, hsp_length
        self.hsp_evalue, self.hsp_bits, self.hsp_identity = hsp_evalue, hsp_bits, hsp_identity
        self.hsp_positives = hsp_positives

    def __str__(self):
        """Simple printing function."""
        line = [self.query, self.target, self.query_hsp_start,
                self.query_hsp_end, self.target_hsp_start,
                self.target_hsp_end, self.hsp_evalue]
        return "\t".join([str(x) for x in line])

    @classmethod
    def from_dict(cls, counter, query_id, target_id, hsp_dict):

        return Hsp(query_id=query_id, target_id=target_id, counter=counter,
                   **hsp_dict)

    @classmethod
    def from_hsp(cls, counter, query_id, target_id, hsp, qmultiplier=1, tmultiplier=1):
        hsp_dict, _, _ = prepare_hsp(hsp, counter, qmultiplier=qmultiplier, tmultiplier=tmultiplier)
        return cls(query_id=query_id, target_id=target_id, counter=counter, **hsp_dict)

    @classmethod
    def as_dict_static(cls, state_obj):
        """Method to return a dict representation of the object. Necessary for storing.
        This method returns a dictionary *without any attribute that requires joined data*.
        As a static method, it is useful to be used outside of the class.
        :param state_obj: an instance of the HSP class or a
        collections.namedtuple object from a direct query

        :rtype : dict
        """

        keys = [
            "query_hsp_start",
            "query_hsp_end",
            "target_hsp_start",
            "target_hsp_end",
            "hsp_evalue",
            "hsp_bits",
            "match",
            "query_frame",
            "target_frame"
        ]

        state = dict().fromkeys(keys)
        for key in keys:
            state[key] = getattr(state_obj, key)
        return state

    def as_dict(self):
        """Method to return a dict representation of the object. Necessary for storing.
        This method returns a dictionary *without any attribute that requires joined data*.
        It is meant to be used only by the method as_dict of the Hit class."""

        return self.as_dict_static(self)

    def as_full_dict(self):
        """Method to return a dict representation of the object.
        This method also checks query name and hit name, so it is slower than as_dict and used
        when it is necessary to retrieve data independently from Hit.
        """

        state = self.as_dict()
        state["query"] = self.query
        state["target"] = self.target
        state["query_hsp_cov"] = self.query_hsp_cov
        state["target_hsp_cov"] = self.target_hsp_cov
        state["match"] = self.match

        return state

    @hybrid_property
    def query(self):
        return self.query_object.query_name

    @hybrid_property
    def query_length(self):
        return self.query_object.query_length

    @hybrid_property
    def target(self):
        return self.target_object.target_name

    @hybrid_property
    def target_length(self):
        return self.target_object.target_length

    @hybrid_property
    def query_hsp_cov(self):
        """This property returns the percentage of the query which is covered by the HSP."""
        val = (self.query_hsp_end - self.query_hsp_start + 1)
        val /= self.query_length
        return val

    @hybrid_property
    def target_hsp_cov(self):
        """This property returns the percentage of the target which is covered by the HSP."""
        val = (self.target_hsp_end - self.target_hsp_start + 1)
        val /= self.target_length
        return val


def prepare_hsp(hsp, counter, off_by_one=False, qmultiplier=1, tmultiplier=1):

    r"""
    Prepare a HSP for loading into the DB.
    The match line will be reworked in the following way:

    - If the position is a match/positive, keep the original value
    - If the position is a gap *for the query*, insert a - (dash)
    - If the position is a gap *for the target*, insert a _ (underscore)
    - If the position is a gap *for both*, insert a \ (backslash)

    :param hsp: An HSP object from Bio.Blast.NCBIXML
    :type hsp: Bio.SearchIO.HSP
    :param counter: a digit that indicates the priority of the HSP in the hit
    :return: hsp_dict, identical_positions, positives
    :rtype: (dict, set, set)
    """

    hsp_dict = dict()
    # We must start from 1, otherwise MySQL crashes as its indices start from 1 not 0
    hsp_dict["counter"] = counter + 1
    hsp_dict["query_hsp_start"] = hsp.query_start
    match, identical_positions, positives = prepare_aln_strings(hsp, off_by_one=off_by_one,
                                                                qmultiplier=qmultiplier)
    hsp_dict["query_hsp_end"] = hsp.query_end - off_by_one
    hsp_dict["query_frame"] = hsp.query_frame
    hsp_dict["target_hsp_start"] = hsp.hit_start
    hsp_dict["target_hsp_end"] = hsp.hit_end
    hsp_dict["target_frame"] = hsp.hit_frame
    hsp_dict["hsp_identity"] = hsp.ident_num / hsp.aln_span * 100
    hsp_dict["hsp_positives"] = (match.count("+") + match.count("|")) / hsp.aln_span * 100
    hsp_dict["match"] = match
    hsp_dict["hsp_length"] = hsp.aln_span
    hsp_dict["hsp_bits"] = hsp.bitscore
    hsp_dict["hsp_evalue"] = hsp.evalue
    return hsp_dict, identical_positions, positives



"""
This module contains the methods related to creating the proper output lines for printing
GFFs/GTFs starting from the transcript class.
"""


from itertools import zip_longest
import functools
from ...parsers.GTF import GtfLine
from ...parsers.GFF import GffLine
from ...parsers.bed12 import BED12
import numpy as np


__author__ = 'Luca Venturini'


gff_constructor = GffLine.string_from_dict
gtf_constructor = GtfLine.string_from_dict


def __create_cds_lines(transcript,
                       cds_run,
                       tid,
                       to_gtf=False,
                       with_introns=False):

    """
    Private method to create the exon/UTR/CDS lines for printing
    out in GTF/GFF format.
    :param transcript: the transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript
    :param cds_run: the internal orf run we are preparing
    :param tid: name of the transcript
    :param to_gtf: boolean, indicates whether the lines should be GTF or GFF
    :param with_introns: boolean, if True introns will be added to the output
    we want GTF or GFF output
    :return:
    """

    exon_lines = []
    cds_begin = False
    counter = dict()

    line_creator = functools.partial(__create_exon_line,
                                     transcript,
                                     **{"to_gtf": to_gtf,
                                        "tid": tid})

    if with_introns is True:
        cds_run = cds_run[:]
        for intron in transcript.introns:
            cds_run.append(("intron", intron))

    cds_run = sorted(cds_run, key=lambda segment: (segment[1][0],
                                                   segment[0].lower()))

    for segment in cds_run:
        try:
            exon_line, counter, cds_begin = line_creator(segment,
                                                         counter,
                                                         cds_begin)
        except IndexError:
            raise IndexError(cds_run)
        exon_lines.append(exon_line)

    # assert not any(True for x in exon_lines if x.feature == "CDS" and x.phase is None), [str(_) for _ in exon_lines]

    return [str(line) for line in exon_lines]


# pylint: disable=too-many-arguments
def __create_exon_line(transcript, segment, counter, cds_begin,
                       tid="", to_gtf=False):
    """
    Private method that creates an exon line for printing.
    :param transcript: the transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param segment: a segment of the form (feature, start, end)
    :type segment: list(str, tuple)

    :param counter: a dict object that keeps track of how many exons,
    CDS, UTR segments we have already seen
    :type counter: dict

    :param cds_begin: boolean flag that indicates whether the CDS has already begun
    :type cds_begin: bool

    :param tid: name of the transcript
    :param to_gtf: boolean flag

    :return: exon_line, counter, cds_begin
    :rtype: str, dict, bool
    """

    if to_gtf is False:
        constructor = gff_constructor
        utr3_feature = "three_prime_UTR"
        utr5_feature = "five_prime_UTR"
    else:
        constructor = gtf_constructor
        utr3_feature = "3UTR"
        utr5_feature = "5UTR"

    assert segment[0] in ("UTR", "CDS", "exon", "intron"), segment

    if hasattr(transcript, "chrom"):
        chrom, source, strand = transcript.chrom, transcript.source, transcript.strand
        parent = transcript.parent
    else:
        chrom, source, strand = transcript["chrom"], transcript["source"], transcript["strand"]
        try:
            parent = transcript["parent"]
        except KeyError:
            raise KeyError((type(transcript), transcript))

    if not source:
        source = "Mikado"
    if not strand:
        strand = "."

    phase = None
    if segment[0] == "UTR":
        if (cds_begin is True and strand == "-") or \
                (strand == "+" and cds_begin is False):
            feature = utr5_feature
            counter["five"] = counter.get("five", 0) + 1
            index = counter["five"]
        else:
            feature = utr3_feature
            counter["three"] = counter.get("three", 0) + 1
            index = counter["three"]
    elif segment[0] == "CDS":
        cds_begin = True
        counter["CDS"] = counter.get("CDS", 0) + 1
        index = counter["CDS"]
        feature = "CDS"
        try:
            phase = segment[2]
        except IndexError:
            raise IndexError(segment)
    else:
        try:
            counter[segment[0]] = counter.get(segment[0], 0) + 1
        except AttributeError:
            raise AttributeError((segment, counter))
        index = counter[segment[0]]
        feature = segment[0]

    if phase is None:
        phase = "."

    data = {
        "chrom": chrom,
        "source": source if source else "Mikado",
        "feature": feature,
        "start": segment[1][0],
        "end": segment[1][1],
        "strand": strand if strand else ".",
        "phase": phase,
        "score": ".",
        "attributes": dict()
    }

    if to_gtf is True:
        # noinspection PyPropertyAccess
        # assert transcript.parent
        assert tid
        exon_line = constructor(data, gene=parent, transcript=tid)
    else:
        exon_line = constructor(data,
                                parent=tid,
                                mid="{0}.{1}{2}".format(tid, feature, index),
                                name=None,
                                attribute_order=None)

    return exon_line, counter, cds_begin
# pylint: enable=too-many-arguments


def create_lines_cds(transcript,
                     to_gtf=False,
                     with_introns=False,
                     all_orfs=False,
                     transcriptomic=False):

    """
    Method to create the GTF/GFF lines for printing in the presence of CDS information.
    WARNING: at the moment, the phase support is disabled.
    :param transcript: the transcript instance
    :type transcript: Mikado.loci.transcript.Transcript

    :param to_gtf: boolean, it indicates whether the output is GTF (True) or GFF3 (False)

    :param with_introns: boolean, if set to True, introns will be printed as well.
    :return:
    """

    if to_gtf is False:
        constructor = gff_constructor
    else:
        constructor = gtf_constructor

    lines = []
    transcript_counter = 0

    if transcript.is_coding is False:
        lines = create_lines_no_cds(transcript, to_gtf=to_gtf, with_introns=with_introns,
                                    transcriptomic=transcriptomic)
    else:
        if all_orfs is True:
            iterable = transcript.internal_orfs
        else:
            iterable = [transcript.selected_internal_orf]

        for index, cds_run in enumerate(iterable):
            transcript.logger.debug("CDS run for %s: %s", transcript.id, cds_run)
            if transcript.number_internal_orfs > 1 and all_orfs is True:
                transcript_counter += 1
                transcript.attributes["derives_from"] = transcript.id
                tid = "{0}.orf{1}".format(transcript.id, transcript_counter)
                if index == transcript.selected_internal_orf_index:
                    transcript.attributes["maximal"] = True
                else:
                    transcript.attributes["maximal"] = False
            else:
                tid = transcript.id
            cds_run = transcript.internal_orfs[index]

            if transcriptomic is False:
                parent_line = dict(
                    (attr, getattr(transcript, attr))
                    for attr in ("chrom", "source", "feature", "start", "end",
                                 "score", "strand", "attributes")
                )
                if parent_line["score"] is None:
                    parent_line["score"] = "."

                parent_line["phase"] = '.'
                if parent_line["source"] is None:
                    parent_line["source"] = "Mikado"
                if parent_line["strand"] is None:
                    parent_line["strand"] = "."

                if to_gtf is True:
                    parent_line["attributes"]["gene_id"] = transcript.parent
                    parent_line["attributes"]["transcript_id"] = tid
                else:
                    parent_line["attributes"]["parent"] = transcript.parent
                    parent_line["attributes"]["ID"] = tid

                parent_line["attributes"]["Name"] = tid  # transcript.name

                exon_lines = __create_cds_lines(transcript,
                                                cds_run,
                                                tid,
                                                to_gtf=to_gtf,
                                                with_introns=with_introns)

                lines.append(constructor(parent_line))
                lines.extend(exon_lines)
            else:
                parent_line = dict(
                    (attr, getattr(transcript, attr))
                    for attr in ["source", "feature", "score", "attributes"]
                )

                parent_line["parent"] = transcript.parent

                if parent_line["score"] is None:
                    parent_line["score"] = "."

                parent_line["chrom"] = transcript.id
                parent_line["start"] = 1
                parent_line["end"] = transcript.cdna_length
                parent_line["strand"] = "+"
                parent_line["phase"] = "."
                if to_gtf is True:
                    parent_line["attributes"]["gene_id"] = transcript.parent
                    parent_line["attributes"]["transcript_id"] = tid
                else:
                    parent_line["attributes"]["parent"] = transcript.parent
                    parent_line["attributes"]["ID"] = tid
                # data["id = tid
                parent_line["attributes"]["Name"] = tid  # transcript.name
                # parent_line.parent = "{}_gene".format(tid)

                if to_gtf is True:
                    lines.append(
                        constructor(parent_line, gene=transcript.parent, transcript=tid)
                    )
                else:
                    lines.append(
                        constructor(parent_line, parent=transcript.parent, mid=tid,
                                    name=transcript.name)
                    )

                new_cds_run = []

                cds = sorted([_ for _ in cds_run if _[0] == "CDS"])
                exon_line = parent_line.copy()
                exon_line["attributes"].pop("Name", None)
                exon_line["attributes"].pop("name", None)
                exon_line["feature"] = "exon"
                exon_line["score"] = "."

                if to_gtf is True:
                    lines.append(
                        constructor(exon_line, gene=transcript.parent, transcript=tid)
                    )
                else:
                    lines.append(
                        constructor(exon_line, parent=transcript.id, mid=tid + ".exon1",
                                    name=transcript.name)
                    )

                if transcript.strand == "+":
                    cds_start = cds[0][1][0]
                    phase = cds[0][2]
                    five_utr = [_[1] for _ in cds_run if _[0] == "UTR" and _[1][1] < cds_start]
                else:
                    cds_start = cds[-1][1][1]
                    phase = cds[-1][2]
                    five_utr = [_[1] for _ in cds_run if _[0] == "UTR" and _[1][0] > cds_start]
                if five_utr:
                    five_utr = sum([_[1] - _[0] + 1 for _ in five_utr])
                else:
                    five_utr = 0
                if five_utr:
                    cds_start = five_utr + 1
                else:
                    cds_start = 1
                cds_end = sum([_[1][1] - _[1][0] + 1 for _ in cds]) + cds_start - 1

                if five_utr:
                    new_cds_run.append(("UTR", (1, five_utr)))
                new_cds_run.append(("CDS", (cds_start, cds_end), phase))
                if cds_end < transcript.cdna_length:
                    new_cds_run.append(("UTR", (cds_end + 1, transcript.cdna_length)))
                exon_lines = __create_cds_lines(parent_line,
                                                new_cds_run,
                                                tid,
                                                to_gtf=to_gtf,
                                                with_introns=with_introns)
                lines.extend(exon_lines)

    return lines


def as_bed12(transcript, transcriptomic=False, with_cds=True):
    """
    Method to create a BED12 object for printing
    :param transcript: Mikado.loci.transcript.Transcript
    :param transcriptomic: boolean switch
    :param with_cds: boolean switch
    :return:
    """

    transcript.finalize()
    bed12 = BED12(table=transcript.codon_table)
    bed12.transcriptomic = False
    bed12.header = False
    bed12.chrom = transcript.chrom
    bed12.start = transcript.start
    bed12.end = transcript.end

    if transcript.is_coding is True and with_cds is True:
        if transcript.strand != "-":
            try:
                phase = transcript.phases[transcript.selected_cds[0]]
            except KeyError:
                raise KeyError((transcript.selected_cds[0], transcript.phases))
        else:
            try:
                phase = transcript.phases[transcript.selected_cds[-1]]
            except KeyError:
                raise KeyError((transcript.selected_cds[-1], transcript.phases))

    bed12.coding = transcript.is_coding
    bed12.name = transcript.id
    bed12.score = transcript.score if transcript.score else 0
    bed12.strand = transcript.strand
    if transcript.is_coding and with_cds is True:
        bed12.coding = True
        first_exon = [_ for _ in transcript.selected_cds if transcript.selected_cds_start in _]
        assert len(first_exon) == 1
        bed12.phase = transcript.phases[first_exon.pop()]
        bed12.thick_start = transcript.selected_cds[0][0]
        bed12.thick_end = transcript.selected_cds[-1][1]
    else:
        bed12.thick_start = bed12.start
        bed12.thick_end = bed12.end
        bed12.coding = False

    for key, val in transcript.attributes.items():
        bed12.attributes[key] = val

    bed12.block_count = transcript.exon_num
    bed12.block_sizes = [exon[1] - exon[0] + 1 for exon in transcript.exons]
    _introns = np.concatenate([np.array([intron[1] - intron[0] + 1 for intron in sorted(transcript.introns)],
                                        dtype=np.int64),
                               np.zeros(1, dtype=np.int64)])
    bed12.block_starts = np.concatenate([np.zeros(1, dtype=np.int64),
                                         (bed12.block_sizes + _introns).cumsum()[:-1]], axis=0)
    assert bed12.block_starts[0] == 0, bed12.block_starts
    if transcriptomic:
        bed12 = bed12.to_transcriptomic(alias=transcript.alias, start_adjustment=False,
                                        coding=(transcript.is_coding and with_cds))
        bed12.chrom = transcript.id

    return bed12


def create_lines_bed(transcript, transcriptomic=False, with_cds=True):

    """
    Method to return the BED12 format of the transcript.
    :param transcript:
    :param transcriptomic: boolean switch
    :param with_cds: boolean switch
    :return:
    """

    return str(as_bed12(transcript, transcriptomic=transcriptomic, with_cds=with_cds))


def create_lines_no_cds(transcript,
                        to_gtf=False,
                        with_introns=False,
                        transcriptomic=False):

    """
    Method to create the GTF/GFF lines for printing in the absence of CDS information.

    :param transcript: the Transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param to_gtf: boolean, it indicates whether the output is GTF (True) or GFF3 (False)
    :type to_gtf: bool
    """

    if to_gtf is True:
        constructor = gtf_constructor
    else:
        constructor = gff_constructor

    if transcriptomic is False:
        parent_line = {
            "chrom": transcript.chrom,
            "source": transcript.source if transcript.source else "Mikado",
            "feature": transcript.feature,
            "start": transcript.start,
            "end": transcript.end,
            "score": transcript.score if transcript.score is not None else ".",
            "strand": transcript.strand if transcript.strand else ".",
            "phase": ".",
            "attributes": transcript.attributes
        }

        if to_gtf is True:
            # This prevents a strange bug when converting TAIR10 files
            parent_line["attributes"]["transcript_id"] = transcript.id
            parent_line["attributes"]["gene_id"] = transcript.parent
        else:
            parent_line["attributes"]["ID"] = transcript.id
            parent_line["attributes"]["Parent"] = transcript.parent

        parent_line["attributes"]["name"] = transcript.name

        lines = [constructor(parent_line)]
        exon_lines = []

        if with_introns is False:
            intron_list = [None] * len(transcript.exons)
        else:
            intron_list = sorted(transcript.introns)
        counter = dict()

        line_creator = functools.partial(__create_exon_line,
                                         transcript,
                                         **{"to_gtf": to_gtf,
                                            "tid": transcript.id,
                                            "cds_begin": False})

        for exon, intron in zip_longest(sorted(transcript.exons),
                                        intron_list):
            exon_line, counter, _ = line_creator(("exon", exon), counter)

            exon_lines.append(str(exon_line))
            if intron is not None:
                intron_line, counter, _ = line_creator(("intron", intron), counter)
                exon_lines.append(str(intron_line))

        lines.extend(exon_lines)
    else:
        parent_line = {
            "chrom": transcript.id,
            "source": transcript.source if transcript.source else "Mikado",
            "feature": transcript.feature,
            "start": 1,
            "end": transcript.cdna_length,
            "score": transcript.score if transcript.score is not None else ".",
            "strand": "+",
            "phase": ".",
            "attributes": transcript.attributes
        }

        if to_gtf is True:
            parent_line["attributes"]["transcript_id"] = transcript.id
            parent_line["attributes"]["gene_id"] = transcript.parent
        else:
            parent_line["attributes"]["ID"] = transcript.id
            parent_line["attributes"]["Parent"] = transcript.parent

        if parent_line["feature"] == "mRNA":
            parent_line["feature"] = "transcript"

        parent_line["attributes"]["Name"] = transcript.id
        lines = [constructor(parent_line)]
        # assert "Name" in lines[0], lines[0]

        exon_line = parent_line.copy()
        exon_line["parent"] = transcript.id
        exon_line["score"] = "."

        line_creator = functools.partial(__create_exon_line,
                                         exon_line,
                                         **{"to_gtf": to_gtf,
                                            "tid": transcript.id,
                                            "cds_begin": False})

        exon_line, _, _ = line_creator(("exon", (1, transcript.cdna_length)), dict())
        lines.append(exon_line)

    return lines



from ...parsers.bed12 import Bed12Parser
import collections
from ...transcripts import Transcript, Gene


def prepare_reference(reference, queue_logger, ref_gff=False,
                      exclude_utr=False, protein_coding=False) -> (dict, collections.defaultdict):

    """
    Method to prepare the data structures that hold the reference
    information for the parsing.
    :param reference:
    :param queue_logger:
    :param ref_gff:
    :return: genes, positions
    :param exclude_utr:
    :param protein_coding:
    """

    # from ..parsers.GFF import GFF3
    # from ..parsers.GTF import GTF
    # from ..parsers.bam_parser import BamParser

    genes = dict()
    positions = collections.defaultdict(dict)
    transcript2gene = dict()

    for row in reference:
        # Assume we are going to use GTF for the moment
        if row.header is True:
            continue
        elif reference.__annot_type__ == Bed12Parser.__annot_type__:
            transcript = Transcript(row, logger=queue_logger, trust_orf=True, accept_undefined_multi=True)
            if transcript.parent:
                gid = transcript.parent[0]
            else:
                transcript.parent = gid = row.id
            transcript2gene[row.id] = gid
            if gid not in genes:
                genes[gid] = Gene(transcript, gid=gid, logger=queue_logger)
            genes[gid].add(transcript)
        elif row.is_gene is True:
            gid = row.id
            genes[gid] = Gene(row, gid=gid, logger=queue_logger)
        elif row.is_transcript is True or (ref_gff is True and row.feature == "match"):
            queue_logger.debug("Transcript\n%s", str(row))
            transcript = Transcript(row, logger=queue_logger, trust_orf=True, accept_undefined_multi=True)
            if row.feature == "match":
                gid = row.id
            else:
                gid = row.gene

            if gid is None:
                queue_logger.warning("No gene ID found for %s, creating a mock one.", row.id)
                row.parent = f"{row.id}.gene"
                gid = row.parent[0]

            transcript2gene[row.id] = gid
            assert gid is not None
            if gid not in genes:
                genes[gid] = Gene(transcript, gid=gid, logger=queue_logger)
            genes[gid].add(transcript)
            assert transcript.id in genes[gid].transcripts
        elif row.is_exon is True:
            if ref_gff is True:
                if "cDNA_match" in row.feature:
                    row.parent = row.id
                    # row.gene = row.id
                    if row.id not in transcript2gene:
                        genes[row.id] = Gene(None, gid=row.id, logger=queue_logger)
                        transcript2gene[row.id] = row.id
                        transcript = Transcript(row, logger=queue_logger, trust_orf=True, accept_undefined_multi=True)
                        genes[row.id].add(transcript)
                found = False
                for transcript in row.transcript:
                    if transcript in transcript2gene:
                        # We have to perform the check because there are some GFFs
                        # e.g. TAIR
                        # where CDSs are defined within a spurious "Protein" feature
                        found = True
                        gid = transcript2gene[transcript]
                        genes[gid][transcript].add_exon(row)
                if found is False:
                    for transcript in row.transcript:
                        if transcript in genes:  # for pseudogenes and the like
                            found = True
                            genes[transcript].add_exon(row)
                if found is False:
                    queue_logger.warn("This feature has no corresponding transcript! %s",
                                      str(row))
            else:
                if row.gene in genes and row.transcript in genes[row.gene].transcripts:
                    genes[row.gene][row.transcript].add_exon(row)
                else:
                    if row.gene not in genes:
                        genes[row.gene] = Gene(None, gid=row.gene, logger=queue_logger)
                    if row.transcript not in genes[row.gene]:
                        transcript = Transcript(row, logger=queue_logger, trust_orf=True, accept_undefined_multi=True)
                        transcript2gene[row.id] = row.gene
                        genes[row.gene].add(transcript)
                    try:
                        genes[row.gene][row.transcript].add_exon(row)
                    except TypeError as exc:
                        queue_logger.critical("Failed with %s", exc)
                        queue_logger.critical("%s, %s", row.transcript, row)
                        raise

    genes, positions = finalize_reference(genes, positions, queue_logger, exclude_utr=exclude_utr,
                                          protein_coding=protein_coding)

    if len(genes) == 0:
        raise KeyError("No genes remained for the reference!")
    return genes, positions


def finalize_reference(genes, positions, queue_logger, exclude_utr=False, protein_coding=False) \
        -> (dict, collections.defaultdict):

    """
:param genes:
:param positions:
:param queue_logger:
:param exclude_utr:
:param protein_coding:
:return:
"""

    non_coding_to_remove = set()
    genes_to_remove = set()
    for gid in genes:
        genes[gid].logger = queue_logger
        genes[gid].finalize(exclude_utr=exclude_utr)
        if len(genes[gid]) == 0:
            genes_to_remove.add(gid)
            continue
        if protein_coding is True:
            to_remove = []
            for tid in genes[gid].transcripts:
                if genes[gid].transcripts[tid].combined_cds_length == 0:
                    to_remove.append(tid)
                    queue_logger.debug("No CDS for %s", tid)
            if len(to_remove) == len(genes[gid].transcripts):
                non_coding_to_remove.add(gid)
                queue_logger.debug("Noncoding gene: %s", gid)
                continue
            elif len(to_remove) > 0:
                for tid in to_remove:
                    genes[gid].remove(tid)
        key = (genes[gid].start, genes[gid].end)
        if key not in positions[genes[gid].chrom]:
            positions[genes[gid].chrom][key] = []
        positions[genes[gid].chrom][key].append(gid)

    for gid in genes_to_remove:
        queue_logger.warn("Removed from reference: %s; error: %s",
                          gid, genes[gid].exception_message)
        del genes[gid]
    for gid in non_coding_to_remove:
        del genes[gid]
    return genes, positions



from ...transcripts.reference_gene import Gene
from ...exceptions import CorruptIndex
from ...utilities.log_utils import create_null_logger
import sqlite3
import rapidjson as json
from time import sleep
import msgpack
import logging
from .indexing import check_index


# Hack to give the rapidjson library this exception class
# This becomes necessary when we happen to have a corrupted index

class GeneDict:

    def __init__(self, dbname: str, logger=None, exclude_utr=False, check=True, protein_coding=False):

        self.__dbname = dbname
        self.__logger = create_null_logger()
        self.logger = logger
        if check is True:
            try:
                check_index(dbname, self.logger)
            except CorruptIndex:
                raise
        self.__exclude_utr = exclude_utr
        self.__protein_coding = protein_coding
        self.__db = sqlite3.connect("file:{}?mode=ro".format(self.__dbname), uri=True)
        self.__cursor = self.__db.cursor()
        self.__cache = dict()

    @property
    def logger(self):
        return self.__logger

    @logger.setter
    def logger(self, logger):
        if logger is None:
            self.__logger = create_null_logger()
        elif not isinstance(logger, logging.Logger):
            raise TypeError("Invalid logger")
        else:
            self.__logger = logger

    @property
    def positions(self):
        return iter(self.__cursor.execute("SELECT chrom, start, end, gid FROM positions"))

    def __getitem__(self, item):

        if item in self.__cache:
            return self.__cache[item]

        failed = 0
        while True:
            try:
                res = self.__cursor.execute("SELECT json from genes where gid=?", (item,)).fetchmany()
                break
            except sqlite3.OperationalError as exc:
                failed += 1
                if failed > 10:
                    raise sqlite3.OperationalError(exc)
                sleep(3)

        if res:
            try:
                gene = self.__load_gene(res[0][0])
                self.__cache[item] = gene
                return gene
            except IndexError:
                raise IndexError(res)
        else:
            return None

    def __load_gene(self, jdict):
        gene = Gene(None, logger=self.logger)
        if not isinstance(jdict, dict):
            jdict = msgpack.loads(jdict, raw=False)
        gene.load_dict(jdict, exclude_utr=self.__exclude_utr, protein_coding=self.__protein_coding)
        gene.finalize()
        return gene

    def __iter__(self):
        self.__db.row_factory = lambda cursor, row: row[0]
        cursor = self.__db.cursor()
        for gid in cursor.execute("SELECT gid from genes"):
            yield gid
        self.__db.row_factory = None

    def items(self):

        for gid in self:
            yield (gid, self[gid])



#!/usr/bin/env python3

"""
This subprogram is my version of ParsEval/CuffCompare. It produces
class codes for each prediction transcript, like cuffcompare,
and its output statistics are blatantly copied by that program;
however, it is more affine to ParsEval in the sense that its only
purpose is to compare two sets of annotations, with no interest in
assembling or checking the expression.
"""

import argparse
from ..scales.compare import compare
from ..parsers import parser_factory
from multiprocessing import cpu_count


__author__ = "Luca Venturini"


def compare_parser():
    """
    The parser for the comparison function

    :return: the argument parser
    """

    def get_procs(arg):

        return max(min(int(arg), cpu_count()), 1)

    parser = argparse.ArgumentParser(
        'Tool to define the spec/sens of predictions vs. references.')
    input_files = parser.add_argument_group(
        'Prediction and annotation files.')
    input_files.add_argument('-r', '--reference',
                             type=parser_factory,
                             required=True,
                             help="""Reference annotation file.
                             By default, an index will be crated and saved with the suffix
                             ".midx".""")
    targets = input_files.add_mutually_exclusive_group(required=True)
    targets.add_argument('-p', '--prediction',
                         type=parser_factory,
                         help='Prediction annotation file.')
    targets.add_argument("--self", default=False,
                         action="store_true",
                         help="""Flag. If set, the reference will be compared with itself. \
Useful for understanding how the reference transcripts interact with each other. If this option is selected,
the stats file will not be produced.""")
    targets.add_argument("--internal", default=False,
                         action="store_true",
                         help="""Flag. If set, for each gene with more than one transcript isoform
                         each will be compared to the others. Useful for understanding the structural
                         relationships between the transcripts in each gene.""")
    targets.add_argument("--index", default=False,
                         action="store_true",
                         help="""Flag. If set, compare will stop after
                         having generated the GFF index for the reference.""")
    shm = parser.add_mutually_exclusive_group()
    shm.add_argument("--no-shm", action="store_false", default=False, dest="shm",
                     help="Switch off /dev/shm usage.")
    shm.add_argument("--shm", action="store_true", default=False, dest="shm",
                     help="Switch on /dev/shm usage.")
    parser.add_argument('--distance', type=int, default=2000,
                        help='''Maximum distance for a transcript to be considered
                        a polymerase run-on. Default: %(default)s''')
    parser.add_argument('-pc', '--protein-coding',
                        dest="protein_coding", action="store_true",
                        default=False,
                        help="""Flag. If set, only transcripts with a CDS
                        (both in reference and prediction) will be considered.""")
    parser.add_argument("-o", "--out", default="mikado_compare", type=str,
                        help="Prefix for the output files. Default: %(default)s")
    parser.add_argument("-fm", "--fuzzy-intron-match", dest="fuzzymatch", default=0,
                        type=int,
                        help="""This parameter controls whether an intron should be considered as "matched" even if
                        its splices are within N bases from the annotated splice junctions. By default this is set to 0
                        (ie only proper matches count).""")
    parser.add_argument("--lenient", action="store_true", default=False,
                        help="""If set, exonic statistics will be calculated leniently
in the TMAP as well - ie they will consider an exon as match even if only
the internal junction has been recovered.""")
    parser.add_argument("-nF", "--do-not-report-fusions", dest="report_fusions", action="store_false",
                        default=True,
                        help="Flag. If invoked, Mikado compare will not report fusions in the input. Useful \
when the reference transcripts have not been clustered properly into genes (e.g. a Mikado prepare run).")
    parser.add_argument("-eu", "--exclude-utr", dest="exclude_utr",
                        default=False, action="store_true",
                        help="""Flag. If set, reference and prediction transcripts
                        will be stripped of their UTRs (if they are coding).""")
    parser.add_argument("-n", "--no-index", "--no-save-index", dest="no_save_index",
                        action="store_true", default=False,
                        help="""Unless this flag is set, compare will save an index of the
                        reference to quicken multiple calls.""")
    parser.add_argument("-erm", "--extended-refmap", action="store_true", default=False,
                        dest="extended_refmap",
                        help="""Flag. If set, the RefMap will also contain recall
                        and precision statistics - not just the F1.""")
    parser.add_argument("-upa", "--use-prediction-alias", action="store_true", default=False,
                        dest="use_prediction_alias",
                        help="""Flag. If set, Mikado Compare will use the alias - rather than the transcript ID -
                        to report the results for prediction transcripts in the TMAP and REFMAP files.""")
    parser.add_argument("-l", "--log", default=None, type=str)
    parser.add_argument("-v", "--verbose",
                        action="store_true",
                        default=False)
    parser.add_argument("-z", "--gzip",
                        action="store_true",
                        default=False,
                        help="Flag. If set, TMAP and REFMAP files will be GZipped.")
    parser.add_argument("-x", "--processes", default=1,
                        type=get_procs)
    parser.set_defaults(func=compare)

    return parser


if __name__ == '__main__':
    __args__ = compare_parser().parse_args()
    __args__.func(__args__)



from .parser import Parser
import pysam

from ..exceptions import InvalidParsingFormat


class BamParser(Parser):

    __annot_type__ = "bam"

    def __init__(self, handle):
        self.__closed = False
        try:
            self._handle = pysam.AlignmentFile(handle, "rb", check_sq=False)
        except KeyboardInterrupt:
            raise
        except (ValueError, TypeError):
            raise InvalidParsingFormat("This is not a valid BAM file: {}.".format(handle))

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._handle)

    @property
    def name(self):
        return self._handle.filename.decode()



"""
This module contains the methods used by the Transcript class to split an instance into
multiple transcripts, if the conditions are met (multiple ORFs present and BLAST not
supporting them being part of the same transcript).
"""

from ...configuration.daijin_configuration import DaijinConfiguration
from collections import OrderedDict as SortedDict
import collections
import operator
from ...utilities import IntervalTree
from ...utilities import overlap
from ...exceptions import InvalidTranscript
from ...parsers.blast_utils import merge
from ...parsers.bed12 import BED12
from ...configuration.configuration import MikadoConfiguration


__author__ = 'Luca Venturini'


def check_split_by_blast(transcript, cds_boundaries):

    """
    This method verifies if a transcript with multiple ORFs has support by BLAST to
    NOT split it into its different components.

    The minimal overlap between ORF and HSP is defined inside the JSON at the key
        ["chimera_split"]["blast_params"]["minimal_hsp_overlap"]
    basically, we consider a HSP a hit only if the overlap is over a certain threshold
    and the HSP evalue under a certain threshold.

    The split by CDS can be executed in three different ways - PERMISSIVE, LENIENT, STRINGENT:

    - PERMISSIVE: split if two CDSs do not have hits in common,
    even when one or both do not have a hit at all.
    - STRINGENT: split only if two CDSs have hits and none
    of those is in common between them.
    - LENIENT: split if *both* lack hits, OR *both* have hits and none
    of those is in common.

    :param transcript: the transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript
    :param cds_boundaries:
    :return: cds_boundaries
    :rtype: dict
    """

    # Establish the minimum overlap between an ORF and a BLAST hit to consider it
    # to establish belongingness

    assert isinstance(transcript.configuration, (MikadoConfiguration, DaijinConfiguration))

    minimal_overlap = transcript.configuration.pick.chimera_split.blast_params.minimal_hsp_overlap
    transcript.logger.debug("Minimal overlap when checking for %s: %s",
                            transcript.id, minimal_overlap)

    cds_hit_dict = SortedDict().fromkeys(cds_boundaries.keys())
    for key in cds_hit_dict:
        cds_hit_dict[key] = collections.defaultdict(list)

    # BUG, this is a hacky fix
    if not hasattr(transcript, "blast_hits"):
        transcript.logger.warning(
            "BLAST hits store lost for %s! Creating a mock one to avoid a crash",

            transcript.id)
        transcript.blast_hits = []

    transcript.logger.debug("%s has %d possible hits", transcript.id, len(transcript.blast_hits))

    # Determine for each CDS which are the hits available
    min_eval = transcript.configuration.pick.chimera_split.blast_params.hsp_evalue
    for hit in transcript.blast_hits:
        for hsp in iter(_hsp for _hsp in hit["hsps"] if
                        _hsp["hsp_evalue"] <= min_eval):
            for cds_run in cds_boundaries:
                # If I have a valid hit b/w the CDS region and the hit,
                # add the name to the set
                overlap_threshold = minimal_overlap * (cds_run[1] + 1 - cds_run[0])
                overl = overlap(cds_run, (hsp['query_hsp_start'], hsp['query_hsp_end']))

                if overl >= overlap_threshold:
                    cds_hit_dict[cds_run][(hit["target"], hit["target_length"])].append(hsp)
                    transcript.logger.debug(
                        "Overlap %s passed for %s between %s CDS and %s HSP (threshold %s)",
                        overlap,
                        transcript.id,
                        cds_run,
                        (hsp['query_hsp_start'], hsp['query_hsp_end']),
                        overlap_threshold)
                else:
                    transcript.logger.debug(
                        "Overlap %s rejected for %s between %s CDS and %s HSP (threshold %s)",
                        overlap,
                        transcript.id,
                        cds_run, (hsp['query_hsp_start'], hsp['query_hsp_end']),
                        overlap_threshold)

    transcript.logger.debug("Final cds_hit_dict for %s (length %s): %s", transcript.id, len(cds_hit_dict), cds_hit_dict)

    final_boundaries = SortedDict()
    for boundary in __get_boundaries_from_blast(transcript, cds_boundaries, cds_hit_dict):
        if len(boundary) == 1:
            assert len(boundary[0]) == 2
            boundary = boundary[0]
            final_boundaries[boundary] = cds_boundaries[boundary]
        else:
            nboun = (boundary[0][0], boundary[-1][1])
            final_boundaries[nboun] = []
            for boun in boundary:
                final_boundaries[nboun].extend(cds_boundaries[boun])
    transcript.logger.debug("Final boundaries for %s: %s",
                            transcript.id, final_boundaries)

    cds_boundaries = final_boundaries.copy()
    return cds_boundaries


def check_common_hits(transcript, cds_hits, old_hits):
    """
    This private method verifies whether we have to split a transcript
    if there are hits for both ORFs and some of them refer to the same target.
    To do so, we check whether the two CDS runs actually share at least one HSPs
    (in which case we do NOT want to split); if not, we verify whether the HSPs
    cover a large fraction of the target length. If this is the case, we decide to
    break down the transcript because we are probably in the presence of a tandem
    duplication.

    :param transcript: the transcript instance to analyse
    :type transcript: Mikado.loci_objects.transcript.Transcript

    :param cds_hits:
    :param old_hits:
    :return:
    """

    in_common = set.intersection(set(cds_hits.keys()),
                                 set(old_hits.keys()))
    # We do not have any hit in common
    assert isinstance(transcript.configuration, (MikadoConfiguration, DaijinConfiguration))
    min_overlap_duplication = transcript.configuration.pick.chimera_split.blast_params.min_overlap_duplication
    to_break = True
    for common_hit in in_common:
        old_hsps = old_hits[common_hit]
        cds_hsps = cds_hits[common_hit]
        # First check ... do we have HSPs in common?
        # if len(set.intersection(old_hsps, cds_hsps)) > 0:
        #     to_break = False
        #     break
        old_query_boundaries = IntervalTree.from_tuples(
            [(h["query_hsp_start"], h["query_hsp_end"]) for h in old_hsps])

        # old_query_boundaries = IntervalTree([Interval(h["query_hsp_start"],
        #                                               h["query_hsp_end"])
        #                                      for h in old_hsps])
        # Look for HSPs that span the two ORFs
        if any([len(
                old_query_boundaries.search(new_cds["query_hsp_start"],
                                            new_cds["query_hsp_end"])) > 0]
               for new_cds in cds_hsps):
            to_break = False

        old_target_boundaries = IntervalTree.from_tuples([
            (h["target_hsp_start"], h["target_hsp_end"]) for h in old_hsps])
        for cds_hsp in cds_hsps:
            boundary = (cds_hsp["target_hsp_start"], cds_hsp["target_hsp_end"])
            for target_hit in old_target_boundaries.search(*boundary):
                overlap_fraction = transcript.overlap(boundary,
                                                      target_hit)/common_hit[1]
                transcript.logger.debug("Checking overlap duplication for %s; OF %s, minimum %s",
                                        transcript.id, overlap_fraction, min_overlap_duplication)
                if overlap_fraction >= min_overlap_duplication:
                    to_break = True and to_break
                else:
                    to_break = False
    return to_break


def __get_boundaries_from_blast(transcript, cds_boundaries, cds_hit_dict):

    """
    Private method that calculates the CDS boundaries to keep
    given the blast hits. Called by check_split_by_blast
    :param cds_boundaries:
    :return:
    """
    new_boundaries = []
    leniency = transcript.configuration.pick.chimera_split.blast_params.leniency
    transcript.logger.debug("Leniency for %s: %s", transcript.id, leniency)
    for cds_boundary in cds_boundaries:
        if not new_boundaries:
            new_boundaries.append([cds_boundary])
        else:
            old_boundary = new_boundaries[-1][-1]
            cds_hits = cds_hit_dict[cds_boundary]
            old_hits = cds_hit_dict[old_boundary]
            if len(cds_hits) == len(old_hits) == 0:  # No hit found for either CDS
                # If we are stringent, we DO NOT split
                if leniency == "STRINGENT":
                    new_boundaries[-1].append(cds_boundary)
                else:  # Otherwise, we do split
                    new_boundaries.append([cds_boundary])
            elif min(len(cds_hits), len(old_hits)) == 0:  # We have hits for only one
                # If we are permissive, we split
                if leniency == "PERMISSIVE":
                    new_boundaries.append([cds_boundary])
                else:
                    new_boundaries[-1].append(cds_boundary)
            else:
                if check_common_hits(transcript, cds_hits, old_hits) is True:
                    new_boundaries.append([cds_boundary])
                # We have hits in common
                else:
                    new_boundaries[-1].append(cds_boundary)
        # } # Finish BLAST check
    return new_boundaries


def __split_complex_exon(transcript, exon, texon, sentinel, boundary, invert=False):

    """
    Private method used to split an exon when it is only partially coding,
    :param exon: Exon to be analysed
    :param texon: Transcriptomic coordinates of the exon
    :param sentinel: tuple of boolean flags, it indicates whether there are transcripts on the left
    (first member) and/or on the right (second member) of the current instance.

    :param boundary: Transcriptomic coordinates of the ORF boundary.
    :return:
    """

    left, right = sentinel

    to_discard = None
    new_exon = list(exon)

    if texon[1] == boundary[0]:
        # In this case we have that the exon ends exactly at the end of the
        # UTR, so we have to keep a one-base exon
        if left is False:
            transcript.logger.debug("Appending mixed UTR/CDS 5' exon %s", exon)
        else:
            if transcript.strand == "+":
                # Keep only the LAST base
                to_discard = (exon[0], exon[1]-1)
                new_exon = (exon[1]-1, exon[1])
                texon = (texon[1]-1, texon[1])
                transcript.logger.debug("Appending monobase CDS exon %s (Texon %s)",
                                        new_exon,
                                        texon)
            else:
                # Keep only the FIRST base
                to_discard = (exon[0]+1, exon[1])
                new_exon = (exon[0], exon[0]+1)
                texon = (texon[1]-1, texon[1])
                transcript.logger.debug(
                    "Appending monobase CDS exon %s (Texon %s)",
                    new_exon,
                    texon)
    elif texon[0] == boundary[1]:
        # In this case we have that the exon ends exactly at the end of the
        # CDS, so we have to keep a one-base exon
        if right is False:
            transcript.logger.debug(
                "Appending mixed UTR/CDS right exon %s",
                exon)
        else:
            if transcript.strand == "+":
                # In this case we have to keep only the FIRST base
                to_discard = (exon[0]+1, exon[1])
                new_exon = (exon[0], exon[0]+1)
                texon = (texon[0], texon[0]+1)
                transcript.logger.debug(
                    "Appending monobase CDS exon %s (Texon %s)",
                    new_exon,
                    texon)
            else:
                # In this case we have to keep only the LAST base
                to_discard = (exon[0], exon[1]-1)
                new_exon = (exon[1]-1, exon[1])
                texon = (texon[0], texon[0]+1)
                transcript.logger.debug(
                    "Appending monobase CDS exon %s (Texon %s)",
                    new_exon,
                    texon)

    elif texon[0] <= boundary[0] <= boundary[1] <= texon[1]:
        # Monoexonic
        transcript.logger.debug("Exon %s, case 3.1", exon)

        # if transcript.monoexonic is False:
        if invert is True:
            if left is True:
                new_exon[1] = exon[0] + (texon[1] - boundary[0])
                transcript.logger.debug(
                    "Case 3.1: Negative strand, another transcript on the left, new exon: %d, %d",
                    new_exon[0], new_exon[1])
            if right is True:
                new_exon[0] = exon[1] - (boundary[1] - texon[0])
                transcript.logger.debug(
                    "Case 3.1: Negative strand, another transcript on the right, new exon: %d, %d",
                    new_exon[0], new_exon[1])
        else:
            if left is True:
                new_exon[0] = exon[1] - (texon[1] - boundary[0])
                transcript.logger.debug(
                    "Case 3.1: Positive strand, another transcript on the left, new exon: %d, %d",
                    new_exon[0], new_exon[1])
            if right is True:
                new_exon[1] = exon[0] + (boundary[1] - texon[0])
                transcript.logger.debug(
                    "Case 3.1: Positive strand, another transcript on the right, new exon: %d, %d",
                    new_exon[0], new_exon[1])

        transcript.logger.debug(
            "[Monoexonic] Tstart shifted for %s, %d to %d",
            transcript.id, texon[0], boundary[0])
        transcript.logger.debug(
            "[Monoexonic] GStart shifted for %s, %d to %d",
            transcript.id, exon[0], new_exon[1])
        transcript.logger.debug(
            "[Monoexonic] Tend shifted for %s, %d to %d",
            transcript.id, texon[1], boundary[1])
        transcript.logger.debug(
            "[Monoexonic] Gend shifted for %s, %d to %d",
            transcript.id, exon[1], new_exon[1])

        if left is True:
            texon[0] = boundary[0]
        if right is True:
            texon[1] = boundary[1]

    elif texon[0] <= boundary[0] <= texon[1] <= boundary[1]:
        # In this case we have that exon is sitting halfway
        # i.e. there is a partial 5'UTR
        if left is True:
            if transcript.strand == "-":
                new_exon[1] = exon[0] + (texon[1] - boundary[0])
            else:
                new_exon[0] = exon[1] - (texon[1] - boundary[0])
            transcript.logger.debug(
                "Tstart shifted for %s, %d to %d", transcript.id, texon[0], boundary[0])
            transcript.logger.debug(
                "GStart shifted for %s, %d to %d", transcript.id, exon[0], new_exon[1])
            texon[0] = boundary[0]

    elif texon[1] >= boundary[1] >= texon[0] >= boundary[0]:
        # In this case we have that exon is sitting halfway
        # i.e. there is a partial 3'UTR
        if right is True:
            if transcript.strand == "-":
                new_exon[0] = exon[1] - (boundary[1] - texon[0])
            else:
                new_exon[1] = exon[0] + (boundary[1] - texon[0])
            transcript.logger.debug(
                "Tend shifted for %s, %d to %d",
                transcript.id, texon[1], boundary[1])
            transcript.logger.debug(
                "Gend shifted for %s, %d to %d",
                transcript.id, exon[1], new_exon[1])
            texon[1] = boundary[1]
        else:
            transcript.logger.debug("New exon: %s", new_exon)
            transcript.logger.debug("New texon: %s", texon)

    # Prevent monobase exons
    # if new_exon[0] == new_exon[1]:
    #     new_exon[1] += 1
    new_exon = tuple([new_exon[0], new_exon[1]])

    return new_exon, texon, to_discard


def __create_splitted_exons(transcript, boundary, left, right, orf_strand):

    """
    Given a boundary in transcriptomic coordinates, this method will extract the
    exons retained in the splitted part of the model.

    :param boundary: the *transcriptomic* coordinates of start/end of the ORF(s)
    to be included in the new transcript
    :type boundary: (int,int)

    :param left: boolean flag indicating whether there is another sub-transcript
    to the left of the one we mean to create, irrespective of *genomic* strand
    :type left: bool

    :param left: boolean flag indicating whether there is another sub-transcript
    to the right of the one we mean to create, irrespective of *genomic* strand
    :type right: bool


    :return: my_exons (final exons), discarded_exons (eventual discarded exons),
    tstart (new transcript start), tend (new transcript end)
    :rtype: (list(int,int),list(int,int),int,int)
    """

    my_exons = []

    discarded_exons = []
    tlength = 0
    tstart = float("Inf")
    tend = float("-Inf")

    if transcript.strand == "-":
        reversal = True
    else:
        reversal = False

    transcript.logger.debug("Starting analysis on %s, boundaries %s, left: %s, \
right: %s, reversal: %s",
                            transcript.id, boundary, left, right, reversal)

    for exon in sorted(transcript.exons, key=operator.itemgetter(0), reverse=reversal):
        # Translate into transcript coordinates
        elength = exon[1] - exon[0] + 1
        texon = [tlength + 1, tlength + elength]
        tlength += elength
        transcript.logger.debug("Analysing exon %s [%s] for %s",
                                exon, texon, transcript.id)

        # SIMPLE CASES
        # Exon completely contained in the ORF
        if boundary[0] <= texon[0] < texon[1] <= boundary[1]:
            transcript.logger.debug("Appending CDS exon %s", exon)
            my_exons.append(exon)
        # Exon on the left of the CDS
        elif texon[1] < boundary[0]:
            if left is False:
                transcript.logger.debug("Appending 5'UTR exon %s", exon)
                my_exons.append(exon)
            else:
                transcript.logger.debug("Discarding 5'UTR exon %s", exon)
                discarded_exons.append(exon)
                continue
        elif texon[0] > boundary[1]:
            if right is False:
                transcript.logger.debug("Appending 3'UTR exon %s", exon)
                my_exons.append(exon)
            else:
                transcript.logger.debug("Discarding 3'UTR exon %s", exon)
                discarded_exons.append(exon)
                continue
        else:
            # exon with partial UTR, go to the relative function
            # to handle these complex cases
            assert transcript.strand is not None
            exon, texon, to_discard = __split_complex_exon(
                transcript, exon, texon, (left, right), boundary,
                invert=(transcript.strand != orf_strand))

            my_exons.append(exon)

            if to_discard is not None:
                discarded_exons.append(to_discard)

        tstart = min(tstart, texon[0])
        tend = max(tend, texon[1])

    return my_exons, discarded_exons, tstart, tend


def __create_splitted_transcripts(transcript, cds_boundaries):

    """
    Private method called by split_by_cds to create the various (N>1) transcripts
    that are its output.
    :param cds_boundaries: a list of int tuples, containing the boundaries
     of the new transcripts.
    :return:
    """

    spans = []
    new_transcripts = []

    for counter, (boundary, bed12_objects) in enumerate(
            sorted(cds_boundaries.items(), key=operator.itemgetter(0))):
        new_transcript = transcript.__class__()
        new_transcript.feature = "mRNA"
        for attribute in ["chrom", "source", "original_source", "score", "strand", "attributes"]:
            setattr(new_transcript, attribute, getattr(transcript, attribute))
        # Determine which ORFs I have on my right and left
        new_transcript.parent = transcript.parent
        left = True
        right = True
        if counter == 0:  # leftmost
            left = False
        if 1 + counter == len(cds_boundaries):  # rightmost
            right = False
        counter += 1  # Otherwise they start from 0
        new_transcript.id = "{0}.split{1}".format(transcript.id, counter)
        new_transcript.external_scores.update(transcript.external_scores.items())
        new_transcript.logger = transcript.logger
        bed12_strand = set(_.strand for _ in bed12_objects)
        assert len(bed12_strand) == 1
        bed12_strand = bed12_strand.pop()

        transcript.logger.debug("Splitting exons for %s", new_transcript.id)
        my_exons, discarded_exons, tstart, tend = __create_splitted_exons(
            transcript, boundary, left, right, bed12_strand)

        transcript.logger.debug("""TID %s counter %d, boundary %s, left %s right %s""",
                                transcript.id,
                                counter,
                                boundary,
                                left,
                                right)

        if right is True:
            transcript.logger.debug("TID %s TEND %d Boun[1] %s",
                                    transcript.id, tend, boundary[1])
        if left is True:
            transcript.logger.debug("TID %s TSTART %d Boun[0] %s",
                                    transcript.id, tstart, boundary[0])

        assert len(my_exons) > 0, (discarded_exons, boundary)

        new_transcript.exons = my_exons

        new_transcript.start = min(exon[0] for exon in new_transcript.exons)
        new_transcript.end = max(exon[1] for exon in new_transcript.exons)
        assert new_transcript.end <= transcript.end
        assert new_transcript.start >= transcript.start
        assert new_transcript.is_coding is False
        new_transcript.configuration = transcript.configuration
        # Now we have to modify the BED12s to reflect
        # the fact that we are starting/ending earlier
        new_transcript.finalize()
        # if transcript.monoexonic is True:
        # if new_transcript.monoexonic is True:
        #     new_transcript.strand = None

        transcript.logger.debug("Relocating %d ORFs into the new transcript (%d, %d), \
tcoordinates (%d, %d)",
                                len(bed12_objects),
                                new_transcript.start, new_transcript.end,
                                tstart, tend
                                )
        new_bed12s = __relocate_orfs(transcript, bed12_objects, tstart, tend)
        assert len([_ for _ in new_bed12s if _.strand == "+"]) > 0
        transcript.logger.debug("Loading %d ORFs into the new transcript (%d, %d): %s",
                                len(new_bed12s),
                                new_transcript.start, new_transcript.end,
                                "\n\t"+"\n".join([str(_) for _ in new_bed12s]))
        new_transcript.logger = transcript.logger
        new_transcript.load_orfs(new_bed12s)

        if new_transcript.selected_cds_length <= 0:
            err_message = "No CDS information retained for {0} split {1}\n".format(
                transcript.id, counter)
            err_message += "BED: {0}".format("\n\t".join([str(x) for x in new_bed12s]))
            raise InvalidTranscript(err_message)

        # Load the blast hits
        __load_blast_hits(new_transcript, boundary, transcript)
        new_transcript.finalize()
        new_transcripts.append(new_transcript)
        nspan = (new_transcript.start, new_transcript.end)
        transcript.logger.debug(
            "Transcript {0} split {1}, discarded exons: {2}".format(
                transcript.id, counter, discarded_exons))
        __check_collisions(transcript, nspan, spans)
        spans.append([new_transcript.start, new_transcript.end])

    return new_transcripts


def __load_blast_hits(new_transcript, boundary, transcript):

    """
    Function to load the BLAST hits into the new splitted transcript.
    :param new_transcript: the splitted transcript
    :type new_transcript: Mikado.loci_objects.Transcript
    :param boundary: tuple(start, end) of the boundary of the new transcript
    :type boundary: tuple(int, int)
    :param transcript:  the original transcript
    :type transcript: Mikado.loci_objects.Transcript
    :return:
    """

    for hit in transcript.blast_hits:
        if overlap((hit["query_start"], hit["query_end"]), boundary) > 0:

            minimal_overlap = transcript.configuration.pick.chimera_split.blast_params.minimal_hsp_overlap
            new_hit = __recalculate_hit(hit, boundary, minimal_overlap)
            if new_hit is not None:
                transcript.logger.debug("""Hit %s,
                                        previous id/query_al_length/t_al_length %f/%f/%f,
                                        novel %f/%f/%f""",
                                        new_hit["target"],
                                        hit["global_identity"],
                                        hit["query_aligned_length"],
                                        hit["target_aligned_length"],
                                        new_hit["global_identity"],
                                        new_hit["query_aligned_length"],
                                        new_hit["target_aligned_length"])

                new_transcript.blast_hits.append(new_hit)
            else:
                transcript.logger.debug("Hit %s did not pass overlap checks for %s",
                                        hit["target"], new_transcript.id)
        else:
            transcript.logger.debug("Ignoring hit %s as it is not intersecting", hit)
            continue


def __check_collisions(transcript, nspan, spans):

    """
    This method checks whether a new transcript collides with a previously
    defined transcript.
    :param nspan:
    :param spans:
    :return:
    """

    if len(spans) == 0:
        return
    for span in spans:
        overl = overlap(span, nspan)

        transcript.logger.debug(
            "Comparing start-ends for split of %s. SpanA: %s SpanB: %s Overlap: %d",
            transcript.id, span,
            nspan, overl)

        if overl > 0:
            err_message = "Invalid overlap for {0}! T1: {1}. T2: {2}".format(
                transcript.id, span, nspan)
            transcript.logger.error(err_message)
            raise InvalidTranscript(err_message)


def __recalculate_hit(hit, boundary, minimal_overlap):
    """Static method to recalculate coverage/identity for new hits."""

    __valid_matches = set([chr(x) for x in range(65, 91)] + [chr(x) for x in range(97, 123)] +
                          ["|"])

    hit_dict = dict()
    for key in iter(k for k in hit.keys() if k not in ("hsps",)):
        hit_dict[key] = hit[key]

    hsp_dict_list = []
    # hit_dict["global_identity"] = []
    q_intervals = []
    t_intervals = []

    identical_positions, positives = set(), set()

    best_hsp = (float("inf"), float("-inf"))

    for hsp in hit["hsps"]:
        _ = overlap((hsp["query_hsp_start"], hsp["query_hsp_end"]), boundary)
        if _ >= minimal_overlap * (boundary[1] + 1 - boundary[0]):
            hsp_dict_list.append(hsp)
            if hsp["hsp_evalue"] < best_hsp[0]:
                best_hsp = (hsp["hsp_evalue"], hsp["hsp_bits"])

            q_intervals.append((hsp["query_hsp_start"], hsp["query_hsp_end"]))
            t_intervals.append((hsp["target_hsp_start"], hsp["target_hsp_end"]))

            query_pos = hsp["query_hsp_start"] - 1

            for amino in hsp["match"]:
                if amino in __valid_matches or amino == "+":
                    query_pos += 1
                    positives.add(query_pos)
                    if amino != "+":
                        identical_positions.add(query_pos)
                elif amino == "_":  # Gap in the target sequence
                    query_pos += 1

    if len(hsp_dict_list) == 0:
        return None

    q_merged_intervals, q_aligned = merge(q_intervals)
    hit_dict["query_aligned_length"] = q_aligned
    hit_dict["query_start"] = q_merged_intervals[0][0]
    hit_dict["query_end"] = q_merged_intervals[-1][1]

    t_merged_intervals, t_aligned = merge(t_intervals)
    hit_dict["target_aligned_length"] = t_aligned
    hit_dict["target_start"] = t_merged_intervals[0][0]
    hit_dict["target_end"] = t_merged_intervals[-1][1]
    hit_dict["global_identity"] = len(identical_positions) * 100 / q_aligned
    hit_dict["global_positives"] = len(positives) * 100 / q_aligned
    hit_dict["hsps"] = hsp_dict_list
    hit_dict["bits"] = max(x["hsp_bits"] for x in hit_dict["hsps"])
    hit_dict["evalue"] = min(x["hsp_evalue"] for x in hit_dict["hsps"])

    return hit_dict


def __relocate_orfs(transcript, bed12_objects, tstart, tend):
    """
    Function to recalculate the coordinates of BED12 objects based on
    the new transcriptomic start/end
    :param bed12_objects: list of the BED12 ORFs to relocate
    :param tstart: New transcriptomic start
    :param tend: New transcriptomic end
    :return:
    """
    new_bed12s = []
    tstart, tend = sorted([tstart, tend])

    for obj in bed12_objects:
        # import copy
        # obj = copy.deepcopy(obj)
        new = BED12(table=transcript.codon_table)
        new.transcriptomic = True
        # Phase is necessary for truncated models
        for attr in ["chrom", "start", "end", "strand", "thick_start", "thick_end",
                     "block_starts", "block_sizes", "phase"]:
            setattr(new, attr, getattr(obj, attr))
        # obj.transcriptomic = True
        if new.strand == "-":
            thick_start = obj.end - obj.thick_end + 1
            thick_end = obj.end - obj.thick_start + 1
            old_start, old_end = tstart, tend
            local_tstart = obj.end - old_end + 1
            local_tend = obj.end - old_start + 1
            assert (old_end - old_start) == (local_tend - local_tstart), (
                (old_start, old_end), (local_tstart, local_tend))
            assert (thick_end - thick_start) == (obj.thick_end - obj.thick_start)
            new.strand = "+"
            new.start = 1
            new.end = local_tend - local_tstart + 1
            new.fasta_length = new.end
            assert new.end > 0 and new.end > new.start
            new.thick_start = max(thick_start - local_tstart + 1, 1)
            assert new.thick_start > 0
            new.thick_end = thick_end - local_tstart + 1
            assert new.thick_end > new.thick_start > 0
            new.block_sizes = [new.end]
            new.block_starts = [new.block_starts]
            transcript.logger.debug("Inverting negative ORF in %s",
                                    transcript.id)
        else:
            new.start = 1
            new.end = min(obj.end, tend) - tstart + 1
            new.fasta_length = new.end
            new.thick_start = min(new.thick_start, tend) - tstart + 1
            new.thick_end = min(obj.thick_end, tend) - tstart + 1
            new.block_sizes = [new.end]
            new.block_starts = new.block_starts[:]

        assert new.thick_start > 0, new.thick_start
        assert new.thick_end > 0, new.thick_end
        assert new.invalid is False, (len(new), new.cds_len, new.fasta_length,
                                      new.invalid_reason,
                                      str(new))

        new_bed12s.append(new)
    assert len(new_bed12s) > 0
    return new_bed12s


def split_by_cds(transcript):
    """This method is used for transcripts that have multiple ORFs.
    It will split them according to the CDS information into multiple transcripts.
    UTR information will be retained only if no ORF is down/upstream.

    :param transcript: the transcript instance
    :type transcript: Mikado.loci_objects.transcript.Transcript
    """

    transcript.finalize()

    # List of the transcript that will be retained

    if transcript.number_internal_orfs < 2:
        new_transcripts = [transcript]  # If we only have one ORF this is easy
    elif transcript.source in transcript.configuration.pick.chimera_split.skip:
        # Disable splitting for transcripts with certain tags
        transcript.logger.debug("%s (label %s) to be skipped for splitting",
                                transcript.id, transcript.id.split("_")[0])
        new_transcripts = [transcript]
    else:
        cds_boundaries = SortedDict()
        for orf in sorted(transcript.loaded_bed12,
                          key=operator.attrgetter("thick_start", "thick_end")):
            cds_boundaries[(orf.thick_start, orf.thick_end)] = [orf]

        # Check whether we have to split or not based on BLAST data
        if transcript.configuration.pick.chimera_split.blast_check is True:
            cds_boundaries = check_split_by_blast(transcript, cds_boundaries)

        if len(cds_boundaries) == 1:
            # Recheck how many boundaries we have - after the BLAST check
            # we might have determined that the transcript has not to be split
            transcript.logger.debug("Not splitting %s", transcript.id)
            new_transcripts = [transcript]
        else:
            try:
                new_transcripts = __create_splitted_transcripts(transcript, cds_boundaries)
            except InvalidTranscript as err:
                exc = InvalidTranscript(err)
                transcript.logger.error("Error in splitting %s by ORF",
                                        transcript.id)
                transcript.logger.exception(exc)
                transcript.logger.error("ORFs: %s",
                                        "\n".join([str(_) for _ in transcript.internal_orfs]))
                transcript.logger.error("BED12: %s",
                                        "\n".join([str(_) for _ in transcript.loaded_bed12]))
                transcript.logger.error("Stripping %s of its CDS.",
                                        transcript.id)
                transcript.strip_cds()
                new_transcripts = [transcript]

    assert len(new_transcripts) > 0, str(transcript)
    __original = set()
    for internal in transcript.internal_orfs:
        __original.add(tuple([_[1] for _ in internal if _[0] == "CDS"]))

    for new_transc in new_transcripts:
        for internal in new_transc.internal_orfs:
            internal = tuple([_[1] for _ in internal if _[0] == "CDS"])
            assert internal in __original, (transcript.id, __original, internal)

        new_transc.verified_introns = set.intersection(set(new_transc.introns),
                                                       transcript.verified_introns)
        new_transc.attributes["split"] = True
        yield new_transc

    return



#!/usr/bin/env python3
# coding: utf_8

"""
Mikado is a Python suite whose purpose is to find and resolve genic loci in a
genomic annotation. This is the library it relies onto.
"""

from .version import __version__

__title__ = "Mikado"
__author__ = 'Luca Venturini'
__license__ = 'GPL3'
__copyright__ = 'Copyright 2015-2020 Luca Venturini'

__all__ = ["configuration",
           "exceptions",
           "loci",
           "parsers",
           "picking",
           "preparation",
           "scales",
           "serializers",
           "subprograms",
           "utilities",
           "__version__"]


from .utilities.log_utils import create_default_logger
from numpy._pytesttester import PytestTester
test = PytestTester(__name__)
